import copy
import os

from PyQt5.QtCore import QDir, Qt, pyqtSlot, QTimer
from PyQt5.QtGui import QIcon, QCloseEvent, QKeySequence
from PyQt5.QtWidgets import (
    QMainWindow,
    QUndoGroup,
    QActionGroup,
    QHeaderView,
    QAction,
    QMessageBox,
    QApplication,
    qApp,
)

from urh import settings, version
from urh.controller.CompareFrameController import CompareFrameController
from urh.controller.GeneratorTabController import GeneratorTabController
from urh.controller.SignalTabController import SignalTabController
from urh.controller.SimulatorTabController import SimulatorTabController
from urh.controller.dialogs.CSVImportDialog import CSVImportDialog
from urh.controller.dialogs.DecoderDialog import DecoderDialog
from urh.controller.dialogs.OptionsDialog import OptionsDialog
from urh.controller.dialogs.ProjectDialog import ProjectDialog
from urh.controller.dialogs.ProtocolSniffDialog import ProtocolSniffDialog
from urh.controller.dialogs.ReceiveDialog import ReceiveDialog
from urh.controller.dialogs.SpectrumDialogController import SpectrumDialogController
from urh.controller.widgets.SignalFrame import SignalFrame
from urh.models.FileFilterProxyModel import FileFilterProxyModel
from urh.models.FileIconProvider import FileIconProvider
from urh.models.FileSystemModel import FileSystemModel
from urh.models.ParticipantLegendListModel import ParticipantLegendListModel
from urh.plugins.PluginManager import PluginManager
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.Signal import Signal
from urh.ui.ui_main import Ui_MainWindow
from urh.util import FileOperator, util
from urh.util.Errors import Errors
from urh.util.Logger import logger
from urh.util.ProjectManager import ProjectManager


class MainController(QMainWindow):
    def __init__(self, *args):
        super().__init__(*args)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)

        util.set_splitter_stylesheet(self.ui.splitter)

        OptionsDialog.write_default_options()

        self.project_save_timer = QTimer()
        self.project_manager = ProjectManager(self)
        self.plugin_manager = PluginManager()
        self.signal_tab_controller = SignalTabController(
            self.project_manager, parent=self.ui.tab_interpretation
        )
        self.ui.tab_interpretation.layout().addWidget(self.signal_tab_controller)
        self.compare_frame_controller = CompareFrameController(
            parent=self.ui.tab_protocol,
            plugin_manager=self.plugin_manager,
            project_manager=self.project_manager,
        )
        self.compare_frame_controller.ui.splitter.setSizes([1, 1000000])

        self.ui.tab_protocol.layout().addWidget(self.compare_frame_controller)

        self.generator_tab_controller = GeneratorTabController(
            self.compare_frame_controller,
            self.project_manager,
            parent=self.ui.tab_generator,
        )

        self.simulator_tab_controller = SimulatorTabController(
            parent=self.ui.tab_simulator,
            compare_frame_controller=self.compare_frame_controller,
            generator_tab_controller=self.generator_tab_controller,
            project_manager=self.project_manager,
        )

        self.ui.tab_simulator.layout().addWidget(self.simulator_tab_controller)

        self.undo_group = QUndoGroup()
        self.undo_group.addStack(self.signal_tab_controller.signal_undo_stack)
        self.undo_group.addStack(self.compare_frame_controller.protocol_undo_stack)
        self.undo_group.addStack(self.generator_tab_controller.generator_undo_stack)
        self.undo_group.setActiveStack(self.signal_tab_controller.signal_undo_stack)

        self.cancel_action = QAction(self.tr("Cancel"), self)
        self.cancel_action.setShortcut(
            QKeySequence.Cancel if hasattr(QKeySequence, "Cancel") else "Esc"
        )
        self.cancel_action.triggered.connect(self.on_cancel_triggered)
        self.cancel_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.cancel_action.setIcon(QIcon.fromTheme("dialog-cancel"))
        self.addAction(self.cancel_action)

        self.ui.actionAuto_detect_new_signals.setChecked(
            settings.read("auto_detect_new_signals", True, bool)
        )

        self.participant_legend_model = ParticipantLegendListModel(
            self.project_manager.participants
        )
        self.ui.listViewParticipants.setModel(self.participant_legend_model)

        gtc = self.generator_tab_controller
        gtc.ui.splitter.setSizes([int(gtc.width() / 0.7), int(gtc.width() / 0.3)])

        self.ui.tab_generator.layout().addWidget(self.generator_tab_controller)

        self.signal_protocol_dict = {}  # type: dict[SignalFrame, ProtocolAnalyzer]

        self.ui.lnEdtTreeFilter.setClearButtonEnabled(True)

        group = QActionGroup(self)
        self.ui.actionFSK.setActionGroup(group)
        self.ui.actionOOK.setActionGroup(group)
        self.ui.actionNone.setActionGroup(group)
        self.ui.actionPSK.setActionGroup(group)

        noise_threshold_setting = settings.read("default_noise_threshold", "automatic")
        noise_threshold_group = QActionGroup(self)
        self.ui.actionAutomaticNoiseThreshold.setActionGroup(noise_threshold_group)
        self.ui.actionAutomaticNoiseThreshold.setChecked(
            noise_threshold_setting == "automatic"
        )
        self.ui.action1NoiseThreshold.setActionGroup(noise_threshold_group)
        self.ui.action1NoiseThreshold.setChecked(noise_threshold_setting == "1")
        self.ui.action5NoiseThreshold.setActionGroup(noise_threshold_group)
        self.ui.action5NoiseThreshold.setChecked(noise_threshold_setting == "5")
        self.ui.action10NoiseThreshold.setActionGroup(noise_threshold_group)
        self.ui.action10NoiseThreshold.setChecked(noise_threshold_setting == "10")
        self.ui.action100NoiseThreshold.setActionGroup(noise_threshold_group)
        self.ui.action100NoiseThreshold.setChecked(noise_threshold_setting == "100")

        self.recentFileActionList = []
        self.create_connects()
        self.init_recent_file_action_list(settings.read("recentFiles", [], list))

        self.filemodel = FileSystemModel(self)
        path = QDir.homePath()

        self.filemodel.setIconProvider(FileIconProvider())
        self.filemodel.setRootPath(path)
        self.file_proxy_model = FileFilterProxyModel(self)
        self.file_proxy_model.setSourceModel(self.filemodel)
        self.ui.fileTree.setModel(self.file_proxy_model)

        self.ui.fileTree.setRootIndex(
            self.file_proxy_model.mapFromSource(self.filemodel.index(path))
        )
        self.ui.fileTree.setToolTip(path)
        self.ui.fileTree.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.ui.fileTree.header().setSectionResizeMode(1, QHeaderView.Stretch)
        self.ui.fileTree.setFocus()

        self.generator_tab_controller.table_model.cfc = self.compare_frame_controller

        self.ui.actionConvert_Folder_to_Project.setEnabled(False)

        undo_action = self.undo_group.createUndoAction(self)
        undo_action.setIcon(QIcon.fromTheme("edit-undo"))
        undo_action.setShortcut(QKeySequence.Undo)
        self.ui.menuEdit.insertAction(self.ui.actionDecoding, undo_action)

        redo_action = self.undo_group.createRedoAction(self)
        redo_action.setIcon(QIcon.fromTheme("edit-redo"))
        redo_action.setShortcut(QKeySequence.Redo)
        self.ui.menuEdit.insertAction(self.ui.actionDecoding, redo_action)
        self.ui.menuEdit.insertSeparator(self.ui.actionDecoding)

        self.ui.actionAbout_Qt.setIcon(
            QIcon(":/qt-project.org/qmessagebox/images/qtlogo-64.png")
        )

        self.__set_non_project_warning_visibility()

        self.ui.splitter.setSizes([0, 1])
        self.refresh_main_menu()

        self.apply_default_view(settings.read("default_view", type=int))
        self.project_save_timer.start(
            ProjectManager.AUTOSAVE_INTERVAL_MINUTES * 60 * 1000
        )

        self.ui.actionProject_settings.setVisible(False)
        self.ui.actionSave_project.setVisible(False)
        self.ui.actionClose_project.setVisible(False)

        self.restoreGeometry(
            settings.read("{}/geometry".format(self.__class__.__name__), type=bytes)
        )

    def __set_non_project_warning_visibility(self):
        show = (
            settings.read("show_non_project_warning", True, bool)
            and not self.project_manager.project_loaded
        )
        self.ui.labelNonProjectMode.setVisible(show)

    def create_connects(self):
        self.ui.actionFullscreen_mode.setShortcut(QKeySequence.FullScreen)
        self.ui.actionOpen.setShortcut(QKeySequence(QKeySequence.Open))
        self.ui.actionOpen_directory.setShortcut(QKeySequence("Ctrl+Shift+O"))

        self.ui.menuEdit.aboutToShow.connect(self.on_edit_menu_about_to_show)

        self.ui.actionNew_Project.triggered.connect(
            self.on_new_project_action_triggered
        )
        self.ui.actionNew_Project.setShortcut(QKeySequence.New)
        self.ui.actionProject_settings.triggered.connect(
            self.on_project_settings_action_triggered
        )
        self.ui.actionSave_project.triggered.connect(self.save_project)
        self.ui.actionClose_project.triggered.connect(self.close_project)
        self.ui.actionExit_URH.triggered.connect(self.close)

        self.ui.actionAbout_AutomaticHacker.triggered.connect(
            self.on_show_about_clicked
        )
        self.ui.actionRecord.triggered.connect(
            self.on_show_record_dialog_action_triggered
        )

        self.ui.actionFullscreen_mode.triggered.connect(
            self.on_fullscreen_action_triggered
        )
        self.ui.actionSaveAllSignals.triggered.connect(
            self.signal_tab_controller.save_all
        )
        self.ui.actionCloseAllFiles.triggered.connect(
            self.on_close_all_files_action_triggered
        )
        self.ui.actionOpen.triggered.connect(self.on_open_file_action_triggered)
        self.ui.actionOpen_directory.triggered.connect(
            self.on_open_directory_action_triggered
        )
        self.ui.actionDecoding.triggered.connect(self.on_show_decoding_dialog_triggered)
        self.ui.actionSpectrum_Analyzer.triggered.connect(
            self.on_show_spectrum_dialog_action_triggered
        )
        self.ui.actionOptions.triggered.connect(
            self.show_options_dialog_action_triggered
        )
        self.ui.actionSniff_protocol.triggered.connect(self.show_proto_sniff_dialog)
        self.ui.actionAbout_Qt.triggered.connect(QApplication.instance().aboutQt)
        self.ui.actionSamples_from_csv.triggered.connect(
            self.on_import_samples_from_csv_action_triggered
        )
        self.ui.actionAuto_detect_new_signals.triggered.connect(
            self.on_auto_detect_new_signals_action_triggered
        )

        self.ui.actionAutomaticNoiseThreshold.triggered.connect(
            self.on_action_automatic_noise_threshold_triggered
        )
        self.ui.action1NoiseThreshold.triggered.connect(
            self.on_action_1_noise_threshold_triggered
        )
        self.ui.action5NoiseThreshold.triggered.connect(
            self.on_action_5_noise_threshold_triggered
        )
        self.ui.action10NoiseThreshold.triggered.connect(
            self.on_action_10_noise_threshold_triggered
        )
        self.ui.action100NoiseThreshold.triggered.connect(
            self.on_action_100_noise_threshold_triggered
        )

        self.ui.btnFileTreeGoUp.clicked.connect(self.on_btn_file_tree_go_up_clicked)
        self.ui.fileTree.directory_open_wanted.connect(
            self.project_manager.set_project_folder
        )

        self.signal_tab_controller.frame_closed.connect(self.close_signal_frame)
        self.signal_tab_controller.signal_created.connect(self.on_signal_created)
        self.signal_tab_controller.ui.scrollArea.files_dropped.connect(
            self.on_files_dropped
        )
        self.signal_tab_controller.files_dropped.connect(self.on_files_dropped)
        self.signal_tab_controller.frame_was_dropped.connect(self.set_frame_numbers)

        self.simulator_tab_controller.open_in_analysis_requested.connect(
            self.on_simulator_open_in_analysis_requested
        )
        self.simulator_tab_controller.rx_file_saved.connect(
            self.adjust_for_current_file
        )

        self.compare_frame_controller.show_interpretation_clicked.connect(
            self.show_protocol_selection_in_interpretation
        )
        self.compare_frame_controller.files_dropped.connect(self.on_files_dropped)
        self.compare_frame_controller.show_decoding_clicked.connect(
            self.on_show_decoding_dialog_triggered
        )
        self.compare_frame_controller.ui.treeViewProtocols.files_dropped_on_group.connect(
            self.on_files_dropped_on_group
        )
        self.compare_frame_controller.participant_changed.connect(
            self.signal_tab_controller.on_participant_changed
        )
        self.compare_frame_controller.ui.treeViewProtocols.close_wanted.connect(
            self.on_cfc_close_wanted
        )
        self.compare_frame_controller.show_config_field_types_triggered.connect(
            self.on_show_field_types_config_action_triggered
        )

        self.compare_frame_controller.load_protocol_clicked.connect(
            self.on_compare_frame_controller_load_protocol_clicked
        )
        self.compare_frame_controller.ui.listViewParticipants.doubleClicked.connect(
            self.on_project_settings_action_triggered
        )

        self.ui.lnEdtTreeFilter.textChanged.connect(
            self.on_file_tree_filter_text_changed
        )

        self.ui.tabWidget.currentChanged.connect(self.on_selected_tab_changed)
        self.project_save_timer.timeout.connect(self.save_project)

        self.ui.actionConvert_Folder_to_Project.triggered.connect(
            self.project_manager.convert_folder_to_project
        )
        self.project_manager.project_loaded_status_changed.connect(
            self.on_project_loaded_status_changed
        )
        self.project_manager.project_updated.connect(self.on_project_updated)

        self.ui.textEditProjectDescription.textChanged.connect(
            self.on_text_edit_project_description_text_changed
        )
        self.ui.tabWidget_Project.tabBarDoubleClicked.connect(
            self.on_project_tab_bar_double_clicked
        )

        self.ui.listViewParticipants.doubleClicked.connect(
            self.on_project_settings_action_triggered
        )

        self.ui.actionShowFileTree.triggered.connect(
            self.on_action_show_filetree_triggered
        )
        self.ui.actionShowFileTree.setShortcut(QKeySequence("F10"))

        self.ui.labelNonProjectMode.linkActivated.connect(
            self.on_label_non_project_mode_link_activated
        )

        for i in range(settings.MAX_RECENT_FILE_NR):
            recent_file_action = QAction(self)
            recent_file_action.setVisible(False)
            recent_file_action.triggered.connect(self.on_open_recent_action_triggered)
            self.recentFileActionList.append(recent_file_action)
            self.ui.menuRecent.addAction(self.recentFileActionList[i])

    def add_plain_bits_from_txt(self, filename: str):
        with open(filename) as f:
            protocol = ProtocolAnalyzer.get_protocol_from_string(f.readlines())

        protocol.filename = filename
        protocol.name = util.get_name_from_filename(filename)

        self.compare_frame_controller.add_protocol(protocol)
        self.compare_frame_controller.refresh()
        self.__add_empty_frame_for_filename(protocol, filename)

    def __add_empty_frame_for_filename(self, protocol: ProtocolAnalyzer, filename: str):
        sf = self.signal_tab_controller.add_empty_frame(filename, protocol)
        self.signal_protocol_dict[sf] = protocol
        self.set_frame_numbers()
        self.file_proxy_model.open_files.add(filename)

    def add_protocol_file(self, filename):
        proto = self.compare_frame_controller.add_protocol_from_file(filename)
        if proto:
            self.__add_empty_frame_for_filename(proto, filename)
        self.ui.tabWidget.setCurrentWidget(self.ui.tab_protocol)

    def add_fuzz_profile(self, filename):
        self.ui.tabWidget.setCurrentIndex(2)
        self.generator_tab_controller.load_from_file(filename)

    def add_simulator_profile(self, filename):
        self.ui.tabWidget.setCurrentIndex(3)
        self.simulator_tab_controller.load_simulator_file(filename)

    def add_signalfile(
        self, filename: str, group_id=0, enforce_sample_rate=None, signal_timestamp=0
    ):
        if not os.path.exists(filename):
            QMessageBox.critical(
                self,
                self.tr("File not Found"),
                self.tr(
                    "The file {0} could not be found. Was it moved or renamed?"
                ).format(filename),
            )
            return

        sig_name = os.path.splitext(os.path.basename(filename))[0]

        # Use default sample rate for signal
        # Sample rate will be overridden in case of a project later
        if enforce_sample_rate is not None:
            sample_rate = enforce_sample_rate
        else:
            sample_rate = self.project_manager.device_conf["sample_rate"]

        signal = Signal(
            filename, sig_name, sample_rate=sample_rate, timestamp=signal_timestamp
        )

        self.file_proxy_model.open_files.add(filename)
        self.add_signal(signal, group_id)

    def add_signal(self, signal, group_id=0, index=-1):
        self.setCursor(Qt.WaitCursor)
        pa = ProtocolAnalyzer(signal)
        sig_frame = self.signal_tab_controller.add_signal_frame(pa, index=index)
        pa = self.compare_frame_controller.add_protocol(pa, group_id)

        signal.blockSignals(True)
        has_entry = self.project_manager.read_project_file_for_signal(signal)

        if (
            self.ui.actionAuto_detect_new_signals.isChecked()
            and not has_entry
            and not signal.changed
        ):
            sig_frame.ui.stackedWidget.setCurrentWidget(sig_frame.ui.pageLoading)
            qApp.processEvents()
            if not signal.already_demodulated:
                signal.auto_detect(detect_modulation=True, detect_noise=False)
            sig_frame.ui.stackedWidget.setCurrentWidget(sig_frame.ui.pageSignal)

        signal.blockSignals(False)

        self.signal_protocol_dict[sig_frame] = pa

        sig_frame.refresh_signal(draw_full_signal=True)
        sig_frame.refresh_signal_information(block=True)

        qApp.processEvents()
        sig_frame.show_protocol(refresh=True)

        if self.project_manager.read_participants_for_signal(signal, pa.messages):
            sig_frame.ui.gvSignal.redraw_view()

        sig_frame.ui.gvSignal.auto_fit_view()
        self.set_frame_numbers()

        self.compare_frame_controller.filter_search_results()
        self.refresh_main_menu()
        self.unsetCursor()

    def close_protocol(self, protocol):
        self.compare_frame_controller.remove_protocol(protocol)
        # Needs to be removed in generator also, otherwise program crashes,
        # if item from tree in generator is selected and corresponding signal is closed
        self.generator_tab_controller.tree_model.remove_protocol(protocol)
        protocol.eliminate()

    def close_signal_frame(self, signal_frame: SignalFrame):
        try:
            self.project_manager.write_signal_information_to_project_file(
                signal_frame.signal
            )
            try:
                proto = self.signal_protocol_dict[signal_frame]
            except KeyError:
                proto = None

            if proto is not None:
                self.close_protocol(proto)
                del self.signal_protocol_dict[signal_frame]

            if (
                self.signal_tab_controller.ui.scrlAreaSignals.minimumHeight()
                > signal_frame.height()
            ):
                self.signal_tab_controller.ui.scrlAreaSignals.setMinimumHeight(
                    self.signal_tab_controller.ui.scrlAreaSignals.minimumHeight()
                    - signal_frame.height()
                )

            if signal_frame.signal is not None:
                # Non-Empty Frame (when a signal and not a protocol is opened)
                self.file_proxy_model.open_files.discard(signal_frame.signal.filename)

            signal_frame.eliminate()

            self.compare_frame_controller.ui.treeViewProtocols.expandAll()
            self.set_frame_numbers()
            self.refresh_main_menu()
        except Exception as e:
            Errors.exception(e)
            self.unsetCursor()

    def add_files(self, filepaths, group_id=0, enforce_sample_rate=None):
        num_files = len(filepaths)
        if num_files == 0:
            return

        for i, filename in enumerate(filepaths):
            if not os.path.exists(filename):
                continue

            if os.path.isdir(filename):
                for f in self.signal_tab_controller.signal_frames:
                    self.close_signal_frame(f)

                FileOperator.RECENT_PATH = filename
                self.project_manager.set_project_folder(filename)
                return

            FileOperator.RECENT_PATH = os.path.split(filename)[0]

            if filename.endswith(".complex"):
                self.add_signalfile(
                    filename, group_id, enforce_sample_rate=enforce_sample_rate
                )
            elif filename.endswith(".coco"):
                self.add_signalfile(
                    filename, group_id, enforce_sample_rate=enforce_sample_rate
                )
            elif (
                filename.endswith(".proto")
                or filename.endswith(".proto.xml")
                or filename.endswith(".bin")
            ):
                self.add_protocol_file(filename)
            elif filename.endswith(".wav"):
                try:
                    import wave

                    w = wave.open(filename)
                    w.close()
                except wave.Error as e:
                    Errors.generic_error(
                        "Unsupported WAV type",
                        "Only uncompressed WAVs (PCM) are supported.",
                        str(e),
                    )
                    continue
                self.add_signalfile(
                    filename, group_id, enforce_sample_rate=enforce_sample_rate
                )
            elif filename.endswith(".fuzz") or filename.endswith(".fuzz.xml"):
                self.add_fuzz_profile(filename)
            elif filename.endswith(".sim") or filename.endswith(".sim.xml"):
                self.add_simulator_profile(filename)
            elif filename.endswith(".txt"):
                self.add_plain_bits_from_txt(filename)
            elif filename.endswith(".csv"):
                self.__import_csv(filename, group_id)
                continue
            elif os.path.basename(filename) == settings.PROJECT_FILE:
                self.project_manager.set_project_folder(os.path.split(filename)[0])
            else:
                self.add_signalfile(
                    filename, group_id, enforce_sample_rate=enforce_sample_rate
                )

            if self.project_manager.project_file is None:
                self.adjust_for_current_file(filename)

            self.refresh_main_menu()

    def set_frame_numbers(self):
        self.signal_tab_controller.set_frame_numbers()

    def closeEvent(self, event: QCloseEvent):
        self.save_project()
        settings.write(
            "{}/geometry".format(self.__class__.__name__), self.saveGeometry()
        )
        super().closeEvent(event)

    def close_all_files(self):
        self.signal_tab_controller.close_all()
        self.compare_frame_controller.reset()
        self.generator_tab_controller.table_model.protocol.clear()
        self.generator_tab_controller.refresh_tree()
        self.generator_tab_controller.refresh_table()
        self.generator_tab_controller.refresh_label_list()

        self.signal_tab_controller.signal_undo_stack.clear()
        self.compare_frame_controller.protocol_undo_stack.clear()
        self.generator_tab_controller.generator_undo_stack.clear()

        self.simulator_tab_controller.close_all()

    def show_options_dialog_specific_tab(self, tab_index: int):
        op = OptionsDialog(self.plugin_manager.installed_plugins, parent=self)
        op.values_changed.connect(self.on_options_changed)
        op.ui.tabWidget.setCurrentIndex(tab_index)
        op.show()

    def refresh_main_menu(self):
        enable = len(self.signal_protocol_dict) > 0
        self.ui.actionSaveAllSignals.setEnabled(enable)
        self.ui.actionCloseAllFiles.setEnabled(enable)

    def apply_default_view(self, view_index: int):
        self.compare_frame_controller.ui.cbProtoView.setCurrentIndex(view_index)
        self.generator_tab_controller.ui.cbViewType.setCurrentIndex(view_index)
        self.simulator_tab_controller.ui.cbViewType.setCurrentIndex(view_index)
        for sig_frame in self.signal_tab_controller.signal_frames:
            sig_frame.ui.cbProtoView.setCurrentIndex(view_index)

    def show_project_settings(self):
        pdc = ProjectDialog(
            new_project=False, project_manager=self.project_manager, parent=self
        )
        pdc.finished.connect(self.on_project_dialog_finished)
        pdc.show()

    def collapse_project_tab_bar(self):
        self.ui.tabParticipants.hide()
        self.ui.tabDescription.hide()
        self.ui.tabWidget_Project.setMaximumHeight(
            self.ui.tabWidget_Project.tabBar().height()
        )

    def expand_project_tab_bar(self):
        self.ui.tabDescription.show()
        self.ui.tabParticipants.show()
        self.ui.tabWidget_Project.setMaximumHeight(9000)

    def save_project(self):
        self.project_manager.save_project(
            simulator_config=self.simulator_tab_controller.simulator_config
        )

    def close_project(self):
        self.save_project()
        self.close_all_files()
        self.compare_frame_controller.proto_analyzer.message_types.clear()
        self.compare_frame_controller.active_message_type.clear()
        self.compare_frame_controller.updateUI()
        self.project_manager.participants.clear()
        self.participant_legend_model.update()

        self.filemodel.setRootPath(QDir.homePath())
        self.ui.fileTree.setRootIndex(
            self.file_proxy_model.mapFromSource(self.filemodel.index(QDir.homePath()))
        )
        self.hide_file_tree()

        self.project_manager.project_path = ""
        self.project_manager.project_file = None

    @pyqtSlot()
    def on_project_tab_bar_double_clicked(self):
        if self.ui.tabParticipants.isVisible():
            self.collapse_project_tab_bar()
        else:
            self.expand_project_tab_bar()

    @pyqtSlot()
    def on_project_updated(self):
        self.participant_legend_model.update()
        self.compare_frame_controller.refresh()
        self.ui.textEditProjectDescription.setText(self.project_manager.description)

    @pyqtSlot()
    def on_fullscreen_action_triggered(self):
        if self.ui.actionFullscreen_mode.isChecked():
            self.showFullScreen()
        else:
            self.showMaximized()

    def adjust_for_current_file(self, file_path):
        if file_path is None:
            return

        if file_path in FileOperator.archives.keys():
            file_path = copy.copy(FileOperator.archives[file_path])

        recent_file_paths = settings.read("recentFiles", [], list)
        recent_file_paths = (
            [] if recent_file_paths is None else recent_file_paths
        )  # check None for OSX
        recent_file_paths = [
            p
            for p in recent_file_paths
            if p != file_path and p is not None and os.path.exists(p)
        ]
        recent_file_paths.insert(0, file_path)
        recent_file_paths = recent_file_paths[: settings.MAX_RECENT_FILE_NR]

        self.init_recent_file_action_list(recent_file_paths)

        settings.write("recentFiles", recent_file_paths)

    def init_recent_file_action_list(self, recent_file_paths: list):
        for i in range(len(self.recentFileActionList)):
            self.recentFileActionList[i].setVisible(False)

        if recent_file_paths is None:
            return

        for i, file_path in enumerate(recent_file_paths):
            if os.path.isfile(file_path):
                display_text = os.path.basename(file_path)
                self.recentFileActionList[i].setIcon(QIcon())
            elif os.path.isdir(file_path):
                head, tail = os.path.split(file_path)
                display_text = tail
                head, tail = os.path.split(head)
                if tail:
                    display_text = tail + "/" + display_text

                self.recentFileActionList[i].setIcon(QIcon.fromTheme("folder"))
            else:
                continue

            self.recentFileActionList[i].setText(display_text)
            self.recentFileActionList[i].setData(file_path)
            self.recentFileActionList[i].setVisible(True)

    @pyqtSlot()
    def on_show_field_types_config_action_triggered(self):
        self.show_options_dialog_specific_tab(tab_index=2)

    @pyqtSlot()
    def on_open_recent_action_triggered(self):
        action = self.sender()
        try:
            if os.path.isdir(action.data()):
                self.project_manager.set_project_folder(action.data())
            elif os.path.isfile(action.data()):
                self.setCursor(Qt.WaitCursor)
                self.add_files(
                    FileOperator.uncompress_archives([action.data()], QDir.tempPath())
                )
                self.unsetCursor()
        except Exception as e:
            Errors.exception(e)
            self.unsetCursor()

    @pyqtSlot()
    def on_show_about_clicked(self):
        descr = (
            "<b><h2>Universal Radio Hacker</h2></b>Version: {0}<br />"
            "GitHub: <a href='https://github.com/jopohl/urh'>https://github.com/jopohl/urh</a><br /><br />"
            "Creators:<i><ul><li>"
            "Johannes Pohl &lt;<a href='mailto:joahnnes.pohl90@gmail.com'>johannes.pohl90@gmail.com</a>&gt;</li>"
            "<li>Andreas Noack &lt;<a href='mailto:andreas.noack@hochschule-stralsund.de'>andreas.noack@hochschule-stralsund.de</a>&gt;</li>"
            "</ul></i>".format(version.VERSION)
        )

        QMessageBox.about(self, self.tr("About"), self.tr(descr))

    @pyqtSlot(int, int, int, int)
    def show_protocol_selection_in_interpretation(
        self, start_message, start, end_message, end
    ):
        try:
            cfc = self.compare_frame_controller
            msg_total = 0
            last_sig_frame = None
            for protocol in cfc.protocol_list:
                if not protocol.show:
                    continue
                n = protocol.num_messages
                view_type = cfc.ui.cbProtoView.currentIndex()
                messages = [
                    i - msg_total
                    for i in range(msg_total, msg_total + n)
                    if start_message <= i <= end_message
                ]
                if len(messages) > 0:
                    try:
                        signal_frame = next(
                            (
                                sf
                                for sf, pf in self.signal_protocol_dict.items()
                                if pf == protocol
                            )
                        )
                    except StopIteration:
                        QMessageBox.critical(
                            self,
                            self.tr("Error"),
                            self.tr("Could not find corresponding signal frame."),
                        )
                        return
                    signal_frame.set_roi_from_protocol_analysis(
                        min(messages), start, max(messages), end + 1, view_type
                    )
                    last_sig_frame = signal_frame
                msg_total += n
            focus_frame = last_sig_frame
            if last_sig_frame is not None:
                self.signal_tab_controller.ui.scrollArea.ensureWidgetVisible(
                    last_sig_frame, 0, 0
                )

            QApplication.instance().processEvents()
            self.ui.tabWidget.setCurrentIndex(0)
            if focus_frame is not None:
                focus_frame.ui.txtEdProto.setFocus()
        except Exception as e:
            logger.exception(e)

    @pyqtSlot(str)
    def on_file_tree_filter_text_changed(self, text: str):
        if len(text) > 0:
            self.filemodel.setNameFilters(["*" + text + "*"])
        else:
            self.filemodel.setNameFilters(["*"])

    @pyqtSlot()
    def on_show_decoding_dialog_triggered(self):
        signals = [sf.signal for sf in self.signal_tab_controller.signal_frames]
        decoding_controller = DecoderDialog(
            self.compare_frame_controller.decodings,
            signals,
            self.project_manager,
            parent=self,
        )
        decoding_controller.finished.connect(self.update_decodings)
        decoding_controller.show()
        decoding_controller.decoder_update()

    @pyqtSlot()
    def update_decodings(self):
        self.project_manager.load_decodings()
        self.compare_frame_controller.fill_decoding_combobox()
        self.compare_frame_controller.refresh_existing_encodings()

        self.generator_tab_controller.refresh_existing_encodings(
            self.compare_frame_controller.decodings
        )

    @pyqtSlot(int)
    def on_selected_tab_changed(self, index: int):
        if index == 0:
            self.undo_group.setActiveStack(self.signal_tab_controller.signal_undo_stack)
        elif index == 1:
            self.undo_group.setActiveStack(
                self.compare_frame_controller.protocol_undo_stack
            )
            self.compare_frame_controller.ui.tblViewProtocol.resize_columns()
            self.compare_frame_controller.ui.tblViewProtocol.resize_vertical_header()
            h = max(
                self.compare_frame_controller.ui.btnSaveProto.height(),
                self.generator_tab_controller.ui.btnSave.height(),
            )
            self.compare_frame_controller.ui.btnSaveProto.setMinimumHeight(h)

            th = self.compare_frame_controller.ui.tabWidget.tabBar().height()
            for i in range(self.compare_frame_controller.ui.tabWidget.count()):
                self.compare_frame_controller.ui.tabWidget.widget(
                    i
                ).layout().setContentsMargins(0, 7 + h - th, 0, 0)

        elif index == 2:
            self.undo_group.setActiveStack(
                self.generator_tab_controller.generator_undo_stack
            )
            h = max(
                self.compare_frame_controller.ui.btnSaveProto.height(),
                self.generator_tab_controller.ui.btnSave.height(),
            )
            self.generator_tab_controller.ui.btnSave.setMinimumHeight(h)
            th = self.generator_tab_controller.ui.tabWidget.tabBar().height()
            for i in range(self.generator_tab_controller.ui.tabWidget.count()):
                self.generator_tab_controller.ui.tabWidget.widget(
                    i
                ).layout().setContentsMargins(0, 7 + h - th, 0, 0)
            # Modulators may got changed from Simulator Dialog
            self.generator_tab_controller.refresh_modulators()
            # Signals may got reordered in analysis
            self.generator_tab_controller.tree_model.update()
            self.generator_tab_controller.ui.treeProtocols.expandAll()

    @pyqtSlot()
    def on_show_record_dialog_action_triggered(self):
        pm = self.project_manager
        try:
            r = ReceiveDialog(pm, parent=self)
        except OSError as e:
            logger.error(repr(e))
            return

        if r.has_empty_device_list:
            Errors.no_device()
            r.close()
            return

        r.device_parameters_changed.connect(pm.set_device_parameters)
        r.files_recorded.connect(self.on_signals_recorded)
        r.show()

    def create_protocol_sniff_dialog(self, testing_mode=False):
        pm = self.project_manager
        signal = next(
            (proto.signal for proto in self.compare_frame_controller.protocol_list),
            None,
        )
        signals = [
            f.signal for f in self.signal_tab_controller.signal_frames if f.signal
        ]

        psd = ProtocolSniffDialog(
            project_manager=pm,
            signal=signal,
            signals=signals,
            testing_mode=testing_mode,
            parent=self,
        )

        if psd.has_empty_device_list:
            Errors.no_device()
            psd.close()
            return None
        else:
            psd.device_parameters_changed.connect(pm.set_device_parameters)
            psd.protocol_accepted.connect(
                self.compare_frame_controller.add_sniffed_protocol_messages
            )
            return psd

    @pyqtSlot()
    def show_proto_sniff_dialog(self):
        psd = self.create_protocol_sniff_dialog()
        if psd:
            psd.show()

    @pyqtSlot()
    def on_show_spectrum_dialog_action_triggered(self):
        pm = self.project_manager
        r = SpectrumDialogController(pm, parent=self)
        if r.has_empty_device_list:
            Errors.no_device()
            r.close()
            return

        r.device_parameters_changed.connect(pm.set_device_parameters)
        r.show()

    @pyqtSlot(list)
    def on_signals_recorded(self, recorded_files: list):
        QApplication.instance().setOverrideCursor(Qt.WaitCursor)
        for recorded_file in recorded_files:
            self.add_signalfile(
                recorded_file.filename,
                enforce_sample_rate=recorded_file.sample_rate,
                signal_timestamp=recorded_file.timestamp,
            )
        QApplication.instance().restoreOverrideCursor()

    @pyqtSlot()
    def show_options_dialog_action_triggered(self):
        self.show_options_dialog_specific_tab(tab_index=4)

    @pyqtSlot()
    def on_new_project_action_triggered(self):
        pdc = ProjectDialog(parent=self)
        pdc.finished.connect(self.on_project_dialog_finished)
        pdc.show()

    @pyqtSlot()
    def on_project_settings_action_triggered(self):
        self.show_project_settings()

    @pyqtSlot()
    def on_edit_menu_about_to_show(self):
        self.ui.actionShowFileTree.setChecked(self.ui.splitter.sizes()[0] > 0)

    def hide_file_tree(self):
        self.ui.splitter.setSizes([0, 1])

    @pyqtSlot()
    def on_action_show_filetree_triggered(self):
        if self.ui.splitter.sizes()[0] > 0:
            self.hide_file_tree()
        else:
            self.ui.splitter.setSizes([1, 1])

    @pyqtSlot()
    def on_project_dialog_finished(self):
        if self.sender().committed:
            if self.sender().new_project:
                self.close_project()
                self.project_manager.from_dialog(self.sender())
            else:
                self.project_manager.project_updated.emit()

    @pyqtSlot()
    def on_open_file_action_triggered(self):
        self.show_open_dialog(directory=False)

    @pyqtSlot()
    def on_open_directory_action_triggered(self):
        self.show_open_dialog(directory=True)

    def show_open_dialog(self, directory=False):
        dialog = FileOperator.get_open_dialog(
            directory_mode=directory, parent=self, name_filter="full"
        )
        if dialog.exec_():
            try:
                file_names = dialog.selectedFiles()
                folders = [folder for folder in file_names if os.path.isdir(folder)]

                if len(folders) > 0:
                    folder = folders[0]
                    for f in self.signal_tab_controller.signal_frames:
                        self.close_signal_frame(f)

                    self.project_manager.set_project_folder(folder)
                else:
                    self.setCursor(Qt.WaitCursor)
                    file_names = FileOperator.uncompress_archives(
                        file_names, QDir.tempPath()
                    )
                    self.add_files(file_names)
                    self.unsetCursor()
            except Exception as e:
                Errors.exception(e)
                self.unsetCursor()

    @pyqtSlot()
    def on_close_all_files_action_triggered(self):
        self.close_all_files()

    @pyqtSlot(list)
    def on_files_dropped(self, files):
        """
        :type files: list of QtCore.QUrl
        """
        self.__add_urls_to_group(files, group_id=0)

    @pyqtSlot(list, int)
    def on_files_dropped_on_group(self, files, group_id: int):
        """
        :param group_id:
        :type files: list of QtCore.QUrl
        """
        self.__add_urls_to_group(files, group_id=group_id)

    def __add_urls_to_group(self, file_urls, group_id=0):
        local_files = [
            file_url.toLocalFile() for file_url in file_urls if file_url.isLocalFile()
        ]
        if len(local_files) > 0:
            self.setCursor(Qt.WaitCursor)
            self.add_files(
                FileOperator.uncompress_archives(local_files, QDir.tempPath()),
                group_id=group_id,
            )
            self.unsetCursor()

    @pyqtSlot(list)
    def on_cfc_close_wanted(self, protocols: list):
        frame_protos = {
            sframe: protocol
            for sframe, protocol in self.signal_protocol_dict.items()
            if protocol in protocols
        }

        for frame in frame_protos:
            self.close_signal_frame(frame)

        for proto in (
            proto for proto in protocols if proto not in frame_protos.values()
        ):
            # close protocols without associated signal frame
            self.close_protocol(proto)

    @pyqtSlot(dict)
    def on_options_changed(self, changed_options: dict):
        refresh_protocol_needed = "show_pause_as_time" in changed_options

        if refresh_protocol_needed:
            for sf in self.signal_tab_controller.signal_frames:
                sf.refresh_protocol()

        self.project_manager.reload_field_types()

        self.compare_frame_controller.refresh_field_types_for_labels()
        self.compare_frame_controller.set_shown_protocols()
        self.generator_tab_controller.set_network_sdr_send_button_visibility()
        self.generator_tab_controller.init_rfcat_plugin()
        self.generator_tab_controller.set_modulation_profile_status()
        self.simulator_tab_controller.refresh_field_types_for_labels()
        # Flipper Zero
        self.generator_tab_controller.set_FZSave_button_visibility()

        if "num_sending_repeats" in changed_options:
            self.project_manager.device_conf["num_sending_repeats"] = changed_options[
                "num_sending_repeats"
            ]

        if "default_view" in changed_options:
            self.apply_default_view(int(changed_options["default_view"]))

        if "spectrogram_colormap" in changed_options:
            self.signal_tab_controller.redraw_spectrograms()

    @pyqtSlot()
    def on_text_edit_project_description_text_changed(self):
        self.project_manager.description = (
            self.ui.textEditProjectDescription.toPlainText()
        )

    @pyqtSlot()
    def on_btn_file_tree_go_up_clicked(self):
        cur_dir = self.filemodel.rootDirectory()
        if cur_dir.cdUp():
            path = cur_dir.path()
            self.filemodel.setRootPath(path)
            self.ui.fileTree.setRootIndex(
                self.file_proxy_model.mapFromSource(self.filemodel.index(path))
            )

    @pyqtSlot(int, Signal)
    def on_signal_created(self, index: int, signal: Signal):
        self.add_signal(signal, index=index)

    @pyqtSlot()
    def on_cancel_triggered(self):
        for signal_frame in self.signal_tab_controller.signal_frames:
            signal_frame.cancel_filtering()

    @pyqtSlot()
    def on_import_samples_from_csv_action_triggered(self):
        self.__import_csv(file_name="")

    @pyqtSlot(bool)
    def on_auto_detect_new_signals_action_triggered(self, checked: bool):
        settings.write("auto_detect_new_signals", bool(checked))

    def __import_csv(self, file_name, group_id=0):
        def on_data_imported(complex_file, sample_rate):
            sample_rate = None if sample_rate == 0 else sample_rate
            self.add_files(
                [complex_file], group_id=group_id, enforce_sample_rate=sample_rate
            )

        dialog = CSVImportDialog(file_name, parent=self)
        dialog.data_imported.connect(on_data_imported)
        dialog.exec_()

    @pyqtSlot(str)
    def on_label_non_project_mode_link_activated(self, link: str):
        if link == "dont_show_non_project_again":
            self.ui.labelNonProjectMode.hide()
            settings.write("show_non_project_warning", False)
        elif link == "open_new_project_dialog":
            self.on_new_project_action_triggered()

    @pyqtSlot(bool)
    def on_project_loaded_status_changed(self, project_loaded: bool):
        self.ui.actionProject_settings.setVisible(project_loaded)
        self.ui.actionSave_project.setVisible(project_loaded)
        self.ui.actionClose_project.setVisible(project_loaded)
        self.ui.actionConvert_Folder_to_Project.setDisabled(project_loaded)
        self.__set_non_project_warning_visibility()

    @pyqtSlot()
    def on_compare_frame_controller_load_protocol_clicked(self):
        dialog = FileOperator.get_open_dialog(
            directory_mode=False, parent=self, name_filter="proto"
        )
        if dialog.exec_():
            for filename in dialog.selectedFiles():
                self.add_protocol_file(filename)

    @pyqtSlot(str)
    def on_simulator_open_in_analysis_requested(self, text: str):
        protocol = ProtocolAnalyzer.get_protocol_from_string(text.split("\n"))
        protocol.name = "Transcript"

        self.ui.tabWidget.setCurrentIndex(1)
        self.compare_frame_controller.add_protocol(protocol)
        self.compare_frame_controller.refresh()

    @pyqtSlot()
    def on_action_automatic_noise_threshold_triggered(self):
        settings.write("default_noise_threshold", "automatic")

    @pyqtSlot()
    def on_action_1_noise_threshold_triggered(self):
        settings.write("default_noise_threshold", "1")

    @pyqtSlot()
    def on_action_5_noise_threshold_triggered(self):
        settings.write("default_noise_threshold", "5")

    @pyqtSlot()
    def on_action_10_noise_threshold_triggered(self):
        settings.write("default_noise_threshold", "10")

    @pyqtSlot()
    def on_action_100_noise_threshold_triggered(self):
        settings.write("default_noise_threshold", "100")



import os
import struct
import math

from urh.util.Logger import logger


# Refer to PCAPNG spec
# https://www.ietf.org/staging/draft-tuexen-opsawg-pcapng-02.html


def _build_pcapng_shb(shb_userappl: str = "", shb_hardware: str = "") -> bytes:
    BLOCKTYPE = 0x0A0D0D0A
    HEADERS_BLOCK_LENGTH = 28
    MAGIC_NUMBER = 0x1A2B3C4D
    VERSION_MAJOR, VERSION_MINOR = 1, 0
    SECTIONLENGTH = 0xFFFFFFFFFFFFFFFF  # -1 => Not specified

    shb_userappl_padded_len = math.ceil(len(shb_userappl) / 4) * 4
    shb_hardware_padded_len = math.ceil(len(shb_hardware) / 4) * 4

    total_block_len = HEADERS_BLOCK_LENGTH
    if shb_userappl_padded_len > 0:
        total_block_len += shb_userappl_padded_len + 4

    if shb_hardware_padded_len > 0:
        total_block_len += shb_hardware_padded_len + 4

    shb = struct.pack(
        ">IIIHHQ",
        BLOCKTYPE,
        total_block_len,
        MAGIC_NUMBER,
        VERSION_MAJOR,
        VERSION_MINOR,
        SECTIONLENGTH,
    )

    if shb_userappl != "":
        SHB_USERAPPL = 4
        strpad = shb_userappl.ljust(shb_userappl_padded_len, "\0")
        shb += struct.pack(">HH", SHB_USERAPPL, shb_userappl_padded_len)
        shb += bytes(strpad, "ascii")

    if shb_hardware != "":
        SHB_HARDWARE = 2
        strpad = shb_hardware.ljust(shb_hardware_padded_len, "\0")
        shb += struct.pack(">HH", SHB_HARDWARE, shb_hardware_padded_len)
        shb += bytes(strpad, "ascii")

    shb += struct.pack(">I", total_block_len)
    return shb


def _build_pcapng_idb(link_type) -> bytes:
    BLOCKTYPE = 0x00000001
    BLOCKLENGTH = 20
    SNAP_LEN = 0

    return struct.pack(
        ">IIHHII", BLOCKTYPE, BLOCKLENGTH, link_type, 0, SNAP_LEN, BLOCKLENGTH
    )


def _build_pcapng_epb(packet: bytes, timestamp: float) -> bytes:
    BLOCKTYPE = 0x00000006
    BLOCKHEADERLEN = 32
    INTERFACE_ID = 0

    captured_packet_len = len(packet)
    original_packet_len = captured_packet_len
    padded_packet_len = math.ceil(captured_packet_len / 4) * 4
    padding_len = padded_packet_len - original_packet_len
    padded_packet = packet + bytearray(padding_len)
    block_total_length = BLOCKHEADERLEN + padded_packet_len
    timestamp_int = int(timestamp * 1e6)  # Set the proper resolution
    timestamp_high = timestamp_int >> 32
    timestamp_low = timestamp_int & 0x00000000FFFFFFFF

    epb = struct.pack(
        ">IIIIIII",
        BLOCKTYPE,
        block_total_length,
        INTERFACE_ID,
        timestamp_high,
        timestamp_low,
        captured_packet_len,
        original_packet_len,
    )
    epb += padded_packet
    epb += struct.pack(">I", block_total_length)
    return epb


def create_pcapng_file(
    filename: str, shb_userappl: str = "", shb_hardware: str = "", link_type: int = 147
) -> bytes:
    if filename == "":
        return

    shb_bytes = _build_pcapng_shb(shb_userappl, shb_hardware)
    idb_bytes = _build_pcapng_idb(link_type)

    if os.path.isfile(filename):
        logger.warning("{0} already exists. Overwriting it".format(filename))

    with open(filename, "wb") as f:
        f.write(shb_bytes)
        f.write(idb_bytes)


def append_packets_to_pcapng(filename: str, packets: list, timestamps: list):
    with open(filename, "ab") as f:
        for packet, timestamp in zip(packets, timestamps):
            f.write(_build_pcapng_epb(packet, timestamp))



import os

from PyQt5.QtCore import QModelIndex, pyqtSlot, QFileInfo, pyqtSignal, QUrl
from PyQt5.QtGui import QContextMenuEvent, QIcon, QDesktopServices
from PyQt5.QtWidgets import QTreeView, QInputDialog, QMessageBox, QMenu

from urh import settings
from urh.util import util


class DirectoryTreeView(QTreeView):
    directory_open_wanted = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.doubleClicked.connect(self.on_double_clicked)

    def create_directory(self):
        index = self.model().mapToSource(self.rootIndex())  # type: QModelIndex

        if not index.isValid():
            return

        model = self.model().sourceModel()
        dir_name, ok = QInputDialog.getText(
            self, self.tr("Create Directory"), self.tr("Directory name")
        )

        if ok and len(dir_name) > 0:
            if not model.mkdir(index, dir_name).isValid():
                QMessageBox.information(
                    self,
                    self.tr("Create Directory"),
                    self.tr("Failed to create the directory"),
                )

    def remove(self):
        index = self.model().mapToSource(self.currentIndex())  # type: QModelIndex
        if not index.isValid():
            return

        model = self.model().sourceModel()
        if model.fileInfo(index).isDir():
            ok = model.rmdir(index)
        else:
            ok = model.remove(index)

        if not ok:
            QMessageBox.information(
                self,
                self.tr("Remove"),
                self.tr("Failed to remove {0}".format(model.fileName(index))),
            )

    def create_context_menu(self) -> QMenu:
        menu = QMenu(self)
        index = self.model().mapToSource(self.currentIndex())  # type: QModelIndex
        if index.isValid():
            current_index_info = (
                self.model().sourceModel().fileInfo(index)
            )  # type: QFileInfo
            if current_index_info.isDir():
                if os.path.isfile(
                    os.path.join(current_index_info.filePath(), settings.PROJECT_FILE)
                ):
                    open_action = menu.addAction("Open project")
                    open_action.setIcon(QIcon(":/icons/icons/appicon.png"))
                else:
                    open_action = menu.addAction("Open folder")
                    open_action.setIcon(QIcon.fromTheme("folder-open"))
                open_action.triggered.connect(self.on_open_action_triggered)

        new_dir_action = menu.addAction("New folder")
        new_dir_action.setIcon(QIcon.fromTheme("folder"))
        new_dir_action.triggered.connect(self.create_directory)

        del_action = menu.addAction("Delete")
        del_action.setIcon(QIcon.fromTheme("edit-delete"))
        del_action.triggered.connect(self.remove)

        menu.addSeparator()
        open_in_explorer_action = menu.addAction("Open in file manager...")
        open_in_explorer_action.triggered.connect(
            self.on_open_explorer_action_triggered
        )

        return menu

    def contextMenuEvent(self, event: QContextMenuEvent):
        menu = self.create_context_menu()
        menu.exec_(self.mapToGlobal(event.pos()))

    @pyqtSlot()
    def on_open_action_triggered(self):
        self.directory_open_wanted.emit(self.model().get_file_path(self.currentIndex()))

    @pyqtSlot(QModelIndex)
    def on_double_clicked(self, index: QModelIndex):
        file_path = self.model().get_file_path(index)  # type: str

        if file_path.endswith(".txt"):
            try:
                content = open(file_path, "r").read()
            except:
                return
            d = util.create_textbox_dialog(content, file_path, self)
            d.show()

    @pyqtSlot()
    def on_open_explorer_action_triggered(self):
        file_path = self.model().get_file_path(self.rootIndex())
        QDesktopServices.openUrl(QUrl.fromLocalFile(file_path))



import math
import time
from multiprocessing import Process, Array

import numpy as np
from PyQt5.QtCore import pyqtSignal, QPoint, Qt, QMimeData, pyqtSlot, QTimer
from PyQt5.QtGui import (
    QIcon,
    QDrag,
    QPixmap,
    QRegion,
    QDropEvent,
    QTextCursor,
    QContextMenuEvent,
    QResizeEvent,
)
from PyQt5.QtWidgets import (
    QFrame,
    QMessageBox,
    QMenu,
    QWidget,
    QUndoStack,
    QCheckBox,
    QApplication,
    qApp,
)

from urh import settings
from urh.controller.dialogs.AdvancedModulationOptionsDialog import (
    AdvancedModulationOptionsDialog,
)
from urh.controller.dialogs.CostaOptionsDialog import CostaOptionsDialog
from urh.controller.dialogs.FilterDialog import FilterDialog
from urh.controller.dialogs.SendDialog import SendDialog
from urh.controller.dialogs.SignalDetailsDialog import SignalDetailsDialog
from urh.signalprocessing.Filter import Filter, FilterType
from urh.signalprocessing.IQArray import IQArray
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.Signal import Signal
from urh.signalprocessing.Spectrogram import Spectrogram
from urh.ui.actions.ChangeSignalParameter import ChangeSignalParameter
from urh.ui.actions.EditSignalAction import EditSignalAction, EditAction
from urh.ui.painting.SignalSceneManager import SignalSceneManager
from urh.ui.ui_signal_frame import Ui_SignalFrame
from urh.util import FileOperator, util
from urh.util.Errors import Errors
from urh.util.Formatter import Formatter
from urh.util.Logger import logger


def perform_filter(result_array: Array, data, f_low, f_high, filter_bw):
    result_array = np.frombuffer(result_array.get_obj(), dtype=np.complex64)
    result_array[:] = Filter.apply_bandpass_filter(
        data, f_low, f_high, filter_bw=filter_bw
    )


class SignalFrame(QFrame):
    closed = pyqtSignal(QWidget)
    signal_created = pyqtSignal(Signal)
    drag_started = pyqtSignal(QPoint)
    frame_dropped = pyqtSignal(QPoint)
    files_dropped = pyqtSignal(list)
    not_show_again_changed = pyqtSignal()
    signal_drawing_finished = pyqtSignal()
    apply_to_all_clicked = pyqtSignal(Signal)
    sort_action_clicked = pyqtSignal()

    @property
    def proto_view(self):
        return self.ui.txtEdProto.cur_view

    def __init__(
        self,
        proto_analyzer: ProtocolAnalyzer,
        undo_stack: QUndoStack,
        project_manager,
        parent=None,
    ):
        super().__init__(parent)

        self.undo_stack = undo_stack

        self.ui = Ui_SignalFrame()
        self.ui.setupUi(self)

        util.set_splitter_stylesheet(self.ui.splitter)

        self.__set_spectrogram_adjust_widgets_visibility()
        self.ui.gvSignal.init_undo_stack(self.undo_stack)

        self.ui.txtEdProto.setFont(util.get_monospace_font())
        self.ui.txtEdProto.participants = project_manager.participants
        self.ui.txtEdProto.messages = proto_analyzer.messages

        self.ui.gvSignal.participants = project_manager.participants

        self.filter_abort_wanted = False

        self.setAttribute(Qt.WA_DeleteOnClose)
        self.project_manager = project_manager

        self.proto_analyzer = proto_analyzer
        self.signal = (
            proto_analyzer.signal if self.proto_analyzer is not None else None
        )  # type: Signal
        self.ui.gvSignal.protocol = self.proto_analyzer
        self.ui.gvSignal.set_signal(self.signal)
        self.ui.sliderFFTWindowSize.setValue(
            int(math.log2(Spectrogram.DEFAULT_FFT_WINDOW_SIZE))
        )
        self.ui.sliderSpectrogramMin.setValue(
            self.ui.gvSpectrogram.scene_manager.spectrogram.data_min
        )
        self.ui.sliderSpectrogramMax.setValue(
            self.ui.gvSpectrogram.scene_manager.spectrogram.data_max
        )

        self.dsp_filter = Filter([0.1] * 10, FilterType.moving_average)
        self.set_filter_button_caption()
        self.filter_dialog = FilterDialog(self.dsp_filter, parent=self)

        self.proto_selection_timer = QTimer(self)  # For Update Proto Selection from ROI
        self.proto_selection_timer.setSingleShot(True)
        self.proto_selection_timer.setInterval(1)

        self.spectrogram_update_timer = QTimer(self)
        self.spectrogram_update_timer.setSingleShot(True)
        self.spectrogram_update_timer.setInterval(500)

        # Disabled because never used (see also set_protocol_visibilty())
        self.ui.chkBoxSyncSelection.hide()

        if self.signal is not None:
            self.filter_menu = QMenu()
            self.apply_filter_to_selection_only = self.filter_menu.addAction(
                self.tr("Apply only to selection")
            )
            self.apply_filter_to_selection_only.setCheckable(True)
            self.apply_filter_to_selection_only.setChecked(False)
            self.configure_filter_action = self.filter_menu.addAction(
                "Configure filter..."
            )
            self.configure_filter_action.setIcon(QIcon.fromTheme("configure"))
            self.configure_filter_action.triggered.connect(
                self.on_configure_filter_action_triggered
            )
            self.ui.btnFilter.setMenu(self.filter_menu)

            if not self.signal.already_demodulated:
                self.auto_detect_menu = QMenu()
                self.detect_noise_action = self.auto_detect_menu.addAction(
                    self.tr("Additionally detect noise")
                )
                self.detect_noise_action.setCheckable(True)
                self.detect_noise_action.setChecked(False)
                self.detect_modulation_action = self.auto_detect_menu.addAction(
                    self.tr("Additionally detect modulation")
                )
                self.detect_modulation_action.setCheckable(True)
                self.detect_modulation_action.setChecked(False)
                self.ui.btnAutoDetect.setMenu(self.auto_detect_menu)

            if self.signal.wav_mode:
                if self.signal.already_demodulated:
                    self.ui.lSignalTyp.setText("Demodulated (1-channel *.wav)")
                else:
                    self.ui.lSignalTyp.setText("Signal (*.wav)")
            else:
                self.ui.lSignalTyp.setText("Complex Signal")

            self.ui.lineEditSignalName.setText(self.signal.name)
            self.ui.lSamplesInView.setText("{0:,}".format(self.signal.num_samples))
            self.ui.lSamplesTotal.setText("{0:,}".format(self.signal.num_samples))
            self.sync_protocol = self.ui.chkBoxSyncSelection.isChecked()
            self.ui.chkBoxSyncSelection.hide()

            self.ui.splitter.setSizes([self.ui.splitter.height(), 0])

            self.protocol_selection_is_updateable = True

            self.scene_manager = SignalSceneManager(self.signal, self)
            self.ui.gvSignal.scene_manager = self.scene_manager
            self.scene_manager.scene.setParent(self.ui.gvSignal)
            self.ui.gvSignal.setScene(self.scene_manager.scene)

            self.ui.spinBoxCenterSpacing.setValue(self.signal.center_spacing)
            self.ui.spinBoxBitsPerSymbol.setValue(self.signal.bits_per_symbol)

            self.jump_sync = True
            self.on_btn_show_hide_start_end_clicked()

            self.refresh_signal_information(block=True)
            self.create_connects()
            self.set_protocol_visibility()

            self.ui.chkBoxShowProtocol.setChecked(True)
            self.ui.btnSaveSignal.hide()

            self.show_protocol(refresh=False)

            if self.signal.already_demodulated:
                self.ui.cbModulationType.hide()
                self.ui.labelModulation.hide()
                self.ui.labelNoise.hide()
                self.ui.spinBoxNoiseTreshold.hide()
                self.ui.btnAutoDetect.hide()
                self.ui.cbSignalView.setCurrentIndex(1)
                self.ui.cbSignalView.hide()
                self.ui.lSignalViewText.hide()

        else:
            self.ui.lSignalTyp.setText("Protocol")
            self.set_empty_frame_visibilities()
            self.create_connects()

        self.set_center_spacing_visibility()

    @property
    def spectrogram_is_active(self) -> bool:
        return self.ui.stackedWidget.currentWidget() == self.ui.pageSpectrogram

    def create_connects(self):
        self.ui.btnCloseSignal.clicked.connect(self.on_btn_close_signal_clicked)
        self.ui.btnReplay.clicked.connect(self.on_btn_replay_clicked)
        self.ui.btnAutoDetect.clicked.connect(self.on_btn_autodetect_clicked)
        self.ui.btnInfo.clicked.connect(self.on_info_btn_clicked)
        self.ui.btnShowHideStartEnd.clicked.connect(
            self.on_btn_show_hide_start_end_clicked
        )
        self.filter_dialog.filter_accepted.connect(
            self.on_filter_dialog_filter_accepted
        )
        self.ui.sliderFFTWindowSize.valueChanged.connect(
            self.on_slider_fft_window_size_value_changed
        )
        self.ui.sliderSpectrogramMin.valueChanged.connect(
            self.on_slider_spectrogram_min_value_changed
        )
        self.ui.sliderSpectrogramMax.valueChanged.connect(
            self.on_slider_spectrogram_max_value_changed
        )
        self.ui.gvSpectrogram.y_scale_changed.connect(
            self.on_gv_spectrogram_y_scale_changed
        )
        self.ui.gvSpectrogram.bandpass_filter_triggered.connect(
            self.on_bandpass_filter_triggered
        )
        self.ui.gvSpectrogram.export_fta_wanted.connect(self.on_export_fta_wanted)
        self.ui.btnAdvancedModulationSettings.clicked.connect(
            self.on_btn_advanced_modulation_settings_clicked
        )

        if self.signal is not None:
            self.ui.gvSignal.save_clicked.connect(self.save_signal)

            self.signal.samples_per_symbol_changed.connect(
                self.ui.spinBoxSamplesPerSymbol.setValue
            )
            self.signal.center_changed.connect(self.on_signal_center_changed)
            self.signal.noise_threshold_changed.connect(self.on_noise_threshold_changed)
            self.signal.modulation_type_changed.connect(
                self.ui.cbModulationType.setCurrentText
            )
            self.signal.tolerance_changed.connect(self.ui.spinBoxTolerance.setValue)
            self.signal.protocol_needs_update.connect(self.refresh_protocol)
            self.signal.data_edited.connect(
                self.on_signal_data_edited
            )  # Crop/Delete Mute etc.
            self.signal.bits_per_symbol_changed.connect(
                self.ui.spinBoxBitsPerSymbol.setValue
            )
            self.signal.center_spacing_changed.connect(
                self.on_signal_center_spacing_changed
            )

            self.signal.sample_rate_changed.connect(self.on_signal_sample_rate_changed)

            self.signal.saved_status_changed.connect(
                self.on_signal_data_changed_before_save
            )
            self.ui.btnSaveSignal.clicked.connect(self.save_signal)
            self.signal.name_changed.connect(self.ui.lineEditSignalName.setText)

            self.ui.gvSignal.selection_width_changed.connect(
                self.start_proto_selection_timer
            )
            self.ui.gvSignal.sel_area_start_end_changed.connect(
                self.start_proto_selection_timer
            )
            self.proto_selection_timer.timeout.connect(
                self.update_protocol_selection_from_roi
            )
            self.spectrogram_update_timer.timeout.connect(
                self.on_spectrogram_update_timer_timeout
            )

            self.ui.lineEditSignalName.editingFinished.connect(self.change_signal_name)
            self.proto_analyzer.qt_signals.protocol_updated.connect(
                self.on_protocol_updated
            )

            self.ui.btnFilter.clicked.connect(self.on_btn_filter_clicked)

        self.ui.gvSignal.set_noise_clicked.connect(
            self.on_set_noise_in_graphic_view_clicked
        )
        self.ui.gvSignal.save_as_clicked.connect(self.save_signal_as)
        self.ui.gvSignal.export_demodulated_clicked.connect(self.export_demodulated)

        self.ui.gvSignal.create_clicked.connect(self.create_new_signal)
        self.ui.gvSignal.zoomed.connect(self.on_signal_zoomed)
        self.ui.gvSpectrogram.zoomed.connect(self.on_spectrum_zoomed)
        self.ui.gvSignal.sel_area_start_end_changed.connect(self.update_selection_area)
        self.ui.gvSpectrogram.sel_area_start_end_changed.connect(
            self.update_selection_area
        )
        self.ui.gvSpectrogram.selection_height_changed.connect(
            self.update_number_selected_samples
        )
        self.ui.gvSignal.sep_area_changed.connect(self.set_center)

        self.ui.sliderYScale.valueChanged.connect(self.on_slider_y_scale_value_changed)
        self.ui.spinBoxXZoom.valueChanged.connect(self.on_spinbox_x_zoom_value_changed)

        self.project_manager.project_updated.connect(self.on_participant_changed)
        self.ui.txtEdProto.participant_changed.connect(self.on_participant_changed)
        self.ui.gvSignal.participant_changed.connect(self.on_participant_changed)

        self.proto_selection_timer.timeout.connect(self.update_number_selected_samples)

        self.ui.cbSignalView.currentIndexChanged.connect(
            self.on_cb_signal_view_index_changed
        )
        self.ui.cbModulationType.currentTextChanged.connect(
            self.on_combobox_modulation_type_text_changed
        )
        self.ui.cbProtoView.currentIndexChanged.connect(
            self.on_combo_box_proto_view_index_changed
        )

        self.ui.chkBoxShowProtocol.stateChanged.connect(self.set_protocol_visibility)
        self.ui.chkBoxSyncSelection.stateChanged.connect(
            self.handle_protocol_sync_changed
        )

        self.ui.txtEdProto.proto_view_changed.connect(self.show_protocol)
        self.ui.txtEdProto.show_proto_clicked.connect(
            self.update_roi_from_protocol_selection
        )
        self.ui.txtEdProto.show_proto_clicked.connect(self.zoom_to_roi)
        self.ui.txtEdProto.selectionChanged.connect(
            self.update_roi_from_protocol_selection
        )
        self.ui.txtEdProto.deletion_wanted.connect(
            self.ui.gvSignal.on_delete_action_triggered
        )

        self.ui.spinBoxSelectionStart.valueChanged.connect(
            self.on_spinbox_selection_start_value_changed
        )
        self.ui.spinBoxSelectionEnd.valueChanged.connect(
            self.on_spinbox_selection_end_value_changed
        )
        self.ui.spinBoxCenterOffset.editingFinished.connect(
            self.on_spinbox_center_editing_finished
        )
        self.ui.spinBoxCenterSpacing.valueChanged.connect(
            self.on_spinbox_spacing_value_changed
        )
        self.ui.spinBoxCenterSpacing.editingFinished.connect(
            self.on_spinbox_spacing_editing_finished
        )
        self.ui.spinBoxTolerance.editingFinished.connect(
            self.on_spinbox_tolerance_editing_finished
        )
        self.ui.spinBoxNoiseTreshold.editingFinished.connect(
            self.on_spinbox_noise_threshold_editing_finished
        )
        self.ui.spinBoxSamplesPerSymbol.editingFinished.connect(
            self.on_spinbox_samples_per_symbol_editing_finished
        )
        self.ui.spinBoxBitsPerSymbol.editingFinished.connect(
            self.on_spinbox_bits_per_symbol_editing_finished
        )

    def refresh_signal_information(self, block=True):
        self.ui.spinBoxTolerance.blockSignals(block)
        self.ui.spinBoxCenterOffset.blockSignals(block)
        self.ui.spinBoxSamplesPerSymbol.blockSignals(block)
        self.ui.spinBoxNoiseTreshold.blockSignals(block)
        self.ui.spinBoxBitsPerSymbol.blockSignals(block)
        self.ui.spinBoxCenterSpacing.blockSignals(block)

        self.ui.spinBoxTolerance.setValue(self.signal.tolerance)
        self.ui.spinBoxCenterOffset.setValue(self.signal.center)
        self.ui.spinBoxSamplesPerSymbol.setValue(self.signal.samples_per_symbol)
        self.ui.spinBoxNoiseTreshold.setValue(self.signal.noise_threshold_relative)
        self.ui.cbModulationType.setCurrentText(self.signal.modulation_type)
        self.ui.btnAdvancedModulationSettings.setVisible(
            self.ui.cbModulationType.currentText() in ("ASK", "PSK")
        )
        self.ui.spinBoxCenterSpacing.setValue(self.signal.center_spacing)
        self.ui.spinBoxBitsPerSymbol.setValue(self.signal.bits_per_symbol)

        self.ui.spinBoxTolerance.blockSignals(False)
        self.ui.spinBoxCenterOffset.blockSignals(False)
        self.ui.spinBoxSamplesPerSymbol.blockSignals(False)
        self.ui.spinBoxNoiseTreshold.blockSignals(False)
        self.ui.spinBoxCenterSpacing.blockSignals(False)
        self.ui.spinBoxBitsPerSymbol.blockSignals(False)

        self.set_center_spacing_visibility()

    def set_empty_frame_visibilities(self):
        for widget in dir(self.ui):
            w = getattr(self.ui, widget)
            if hasattr(w, "hide") and w not in (
                self.ui.lSignalNr,
                self.ui.lSignalTyp,
                self.ui.btnCloseSignal,
                self.ui.lineEditSignalName,
            ):
                w.hide()

        self.adjustSize()

    def cancel_filtering(self):
        self.filter_abort_wanted = True

    def update_number_selected_samples(self):
        if self.spectrogram_is_active:
            self.ui.lNumSelectedSamples.setText(
                str(abs(int(self.ui.gvSpectrogram.selection_area.length)))
            )
            self.__set_selected_bandwidth()
            return
        else:
            self.ui.lNumSelectedSamples.setText(
                str(abs(int(self.ui.gvSignal.selection_area.length)))
            )
            self.__set_duration()

        try:
            start, end = int(self.ui.gvSignal.selection_area.start), int(
                self.ui.gvSignal.selection_area.end
            )
            power_str = "-\u221e"  # minus infinity
            if start < end:
                max_window_size = 10**5
                step_size = int(math.ceil((end - start) / max_window_size))
                power = np.mean(
                    self.signal.iq_array.subarray(
                        start, end, step_size
                    ).magnitudes_normalized
                )
                if power > 0:
                    power_str = Formatter.big_value_with_suffix(10 * np.log10(power), 2)

            self.ui.labelRSSI.setText("{} dBm".format(power_str))

        except Exception as e:
            logger.exception(e)
            self.ui.labelRSSI.setText("")

    def change_signal_name(self):
        if self.signal is not None:
            self.signal.name = self.ui.lineEditSignalName.text()

    def __set_spectrogram_adjust_widgets_visibility(self):
        self.ui.labelFFTWindowSize.setVisible(self.ui.cbSignalView.currentIndex() == 2)
        self.ui.sliderFFTWindowSize.setVisible(self.ui.cbSignalView.currentIndex() == 2)
        self.ui.labelSpectrogramMin.setVisible(self.ui.cbSignalView.currentIndex() == 2)
        self.ui.labelSpectrogramMax.setVisible(self.ui.cbSignalView.currentIndex() == 2)
        self.ui.sliderSpectrogramMin.setVisible(
            self.ui.cbSignalView.currentIndex() == 2
        )
        self.ui.sliderSpectrogramMax.setVisible(
            self.ui.cbSignalView.currentIndex() == 2
        )

    def __set_selected_bandwidth(self):
        try:
            num_samples = int(self.ui.lNumSelectedSamples.text())
        except ValueError:
            return

        if self.ui.gvSpectrogram.height_spectrogram and self.signal:
            bw = (
                num_samples / self.ui.gvSpectrogram.height_spectrogram
            ) * self.signal.sample_rate
            self.ui.lDuration.setText(Formatter.big_value_with_suffix(bw) + "Hz")

    def __set_duration(self):  # On Signal Sample Rate changed
        try:
            num_samples = int(self.ui.lNumSelectedSamples.text())
        except ValueError:
            return

        if self.signal:
            t = num_samples / self.signal.sample_rate
            self.ui.lDuration.setText(Formatter.science_time(t))

    def on_slider_y_scale_value_changed(self):
        try:
            gv = (
                self.ui.gvSignal
                if self.ui.stackedWidget.currentIndex() == 0
                else self.ui.gvSpectrogram
            )
            yscale = self.ui.sliderYScale.value()
            current_factor = gv.sceneRect().height() / gv.view_rect().height()
            gv.scale(1, yscale / current_factor)
            x, w = gv.view_rect().x(), gv.view_rect().width()
            gv.centerOn(x + w / 2, gv.y_center)
            if gv.scene_type == 1:
                gv.scene().redraw_legend()
        except ZeroDivisionError:
            pass

    @pyqtSlot()
    def on_slider_fft_window_size_value_changed(self):
        self.spectrogram_update_timer.start()

    @pyqtSlot()
    def on_slider_spectrogram_min_value_changed(self):
        self.spectrogram_update_timer.start()

    @pyqtSlot()
    def on_slider_spectrogram_max_value_changed(self):
        self.spectrogram_update_timer.start()

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            self.drag_started.emit(self.mapToParent(event.pos()))
            drag = QDrag(self)
            mimeData = QMimeData()
            mimeData.setText("Move Signal")
            pixmap = QPixmap(self.rect().size())
            self.render(pixmap, QPoint(), QRegion(self.rect()))
            drag.setPixmap(pixmap)

            drag.setMimeData(mimeData)

            drag.exec_()

    def set_filter_button_caption(self):
        self.ui.btnFilter.setText(
            "Filter ({0})".format(self.dsp_filter.filter_type.value)
        )

    def dragMoveEvent(self, event):
        event.accept()

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent):
        if len(event.mimeData().urls()) == 0:
            self.frame_dropped.emit(self.mapToParent(event.pos()))
        else:
            self.files_dropped.emit(event.mimeData().urls())

    def create_new_signal(self, start, end):
        if start != end:
            new_signal = self.signal.create_new(start=start, end=end)
            self.signal_created.emit(new_signal)
        else:
            Errors.empty_selection()

    def my_close(self):
        not_show = settings.read("not_show_close_dialog", False, type=bool)

        if not not_show:
            cb = QCheckBox("Do not show this again.")
            msgbox = QMessageBox(
                QMessageBox.Question, "Confirm close", "Are you sure you want to close?"
            )
            msgbox.addButton(QMessageBox.Yes)
            msgbox.addButton(QMessageBox.No)
            msgbox.setDefaultButton(QMessageBox.No)
            msgbox.setCheckBox(cb)

            reply = msgbox.exec()

            not_show_again = bool(cb.isChecked())
            settings.write("not_show_close_dialog", not_show_again)
            self.not_show_again_changed.emit()
            if reply != QMessageBox.Yes:
                return

        self.closed.emit(self)

    def save_signal(self):
        if len(self.signal.filename) > 0:
            self.signal.save()
        else:
            self.save_signal_as()

    def save_signal_as(self):
        try:
            FileOperator.ask_signal_file_name_and_save(
                self.signal.name,
                self.signal.iq_array,
                self.signal.sample_rate,
                self.signal.wav_mode,
            )
        except Exception as e:
            Errors.exception(e)

    def export_demodulated(self):
        try:
            initial_name = self.signal.name + "-demodulated.wav"
        except Exception as e:
            logger.exception(e)
            initial_name = "demodulated.wav"

        filename = FileOperator.ask_save_file_name(
            initial_name, caption="Export demodulated"
        )
        if filename:
            try:
                self.setCursor(Qt.WaitCursor)
                data = self.signal.qad
                if filename.endswith(".wav") or filename.endswith(".sub"):
                    data = self.signal.qad.astype(np.float32)
                    data /= np.max(np.abs(data))
                FileOperator.save_data(
                    IQArray(data, skip_conversion=True),
                    filename,
                    self.signal.sample_rate,
                    num_channels=1,
                )
                self.unsetCursor()
            except Exception as e:
                QMessageBox.critical(
                    self, self.tr("Error exporting demodulated data"), e.args[0]
                )

    def draw_signal(self, full_signal=False):
        self.scene_manager.scene_type = self.ui.cbSignalView.currentIndex()
        self.scene_manager.init_scene()
        if full_signal:
            self.ui.gvSignal.show_full_scene()
        else:
            self.ui.gvSignal.redraw_view()

        self.ui.gvSignal.y_sep = -self.signal.center

    def restore_protocol_selection(
        self, sel_start, sel_end, start_message, end_message, old_protoview
    ):
        if old_protoview == self.proto_view:
            return

        self.protocol_selection_is_updateable = False
        sel_start = int(
            self.proto_analyzer.convert_index(
                sel_start, old_protoview, self.proto_view, True
            )[0]
        )
        sel_end = int(
            math.ceil(
                self.proto_analyzer.convert_index(
                    sel_end, old_protoview, self.proto_view, True
                )[1]
            )
        )

        c = self.ui.txtEdProto.textCursor()

        c.setPosition(0)
        cur_message = 0
        i = 0
        text = self.ui.txtEdProto.toPlainText()
        while cur_message < start_message:
            if text[i] == "\n":
                cur_message += 1
            i += 1

        c.movePosition(QTextCursor.Right, QTextCursor.MoveAnchor, i)
        c.movePosition(QTextCursor.Right, QTextCursor.MoveAnchor, sel_start)
        text = text[i:]
        i = 0
        while cur_message < end_message:
            if text[i] == "\n":
                cur_message += 1
            i += 1

        c.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, i)
        c.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, sel_end)

        self.ui.txtEdProto.setTextCursor(c)

        self.protocol_selection_is_updateable = True

    def update_protocol(self):
        self.ui.txtEdProto.setEnabled(False)
        self.ui.txtEdProto.setText("Demodulating...")
        qApp.processEvents()

        try:
            self.proto_analyzer.get_protocol_from_signal()
        except Exception as e:
            Errors.exception(e)

    def show_protocol(self, old_view=-1, refresh=False):
        if not self.proto_analyzer:
            return

        if not self.ui.chkBoxShowProtocol.isChecked():
            if refresh:
                self.update_protocol()
            return

        if old_view == -1:
            old_view = self.ui.cbProtoView.currentIndex()

        if self.proto_analyzer.messages is None or refresh:
            self.update_protocol()
        else:
            # Keep things synchronized and restore selection
            self.ui.txtEdProto.blockSignals(True)
            self.ui.cbProtoView.blockSignals(True)
            self.ui.cbProtoView.setCurrentIndex(self.proto_view)
            self.ui.cbProtoView.blockSignals(False)

            start_message = 0
            sel_start = self.ui.txtEdProto.textCursor().selectionStart()
            text = self.ui.txtEdProto.toPlainText()[:sel_start]
            sel_start = 0
            read_pause = False
            for t in text:
                if t == "\t":
                    read_pause = True

                if not read_pause:
                    sel_start += 1

                if t == "\n":
                    sel_start = 0
                    start_message += 1
                    read_pause = False

            sel_end = self.ui.txtEdProto.textCursor().selectionEnd()
            text = self.ui.txtEdProto.toPlainText()[
                self.ui.txtEdProto.textCursor().selectionStart() : sel_end
            ]
            end_message = 0
            sel_end = 0
            read_pause = False
            for t in text:
                if t == "\t":
                    read_pause = True

                if not read_pause:
                    sel_end += 1

                if t == "\n":
                    sel_end = 0
                    end_message += 1
                    read_pause = False

            self.ui.txtEdProto.setHtml(
                self.proto_analyzer.plain_to_html(self.proto_view)
            )
            try:
                self.restore_protocol_selection(
                    sel_start, sel_end, start_message, end_message, old_view
                )
            except TypeError:
                # Without try/except: segfault (TypeError) when changing sample_rate in info dialog of signal
                pass

            self.ui.txtEdProto.blockSignals(False)

    def draw_spectrogram(self, show_full_scene=False, force_redraw=False):
        self.setCursor(Qt.WaitCursor)
        window_size = 2 ** self.ui.sliderFFTWindowSize.value()
        data_min, data_max = (
            self.ui.sliderSpectrogramMin.value(),
            self.ui.sliderSpectrogramMax.value(),
        )

        redraw_needed = self.ui.gvSpectrogram.scene_manager.set_parameters(
            self.signal.iq_array.data,
            window_size=window_size,
            data_min=data_min,
            data_max=data_max,
        )
        self.ui.gvSpectrogram.scene_manager.update_scene_rect()

        if show_full_scene:
            self.ui.gvSpectrogram.show_full_scene()

        if redraw_needed or force_redraw:
            self.ui.gvSpectrogram.scene_manager.show_full_scene()
            self.ui.gvSpectrogram.show_full_scene()

        self.on_slider_y_scale_value_changed()

        self.__set_samples_in_view()
        self.unsetCursor()

    def eliminate(self):
        self.proto_selection_timer.stop()
        self.ui.verticalLayout.removeItem(self.ui.additionalInfos)

        if self.signal is not None:
            # Avoid memory leaks
            self.scene_manager.eliminate()
            self.signal.eliminate()
            self.proto_analyzer.eliminate()
            self.ui.gvSignal.scene_manager.eliminate()

        self.ui.gvSignal.eliminate()
        self.ui.gvSpectrogram.eliminate()

        self.scene_manager = None
        self.signal = None
        self.proto_analyzer = None

        self.ui.layoutWidget.setParent(None)
        self.ui.layoutWidget.deleteLater()

        self.setParent(None)
        self.deleteLater()

    def __handle_graphic_view_zoomed(self, graphic_view):
        self.ui.lSamplesInView.setText(
            "{0:n}".format(int(graphic_view.view_rect().width()))
        )
        self.ui.spinBoxXZoom.blockSignals(True)
        self.ui.spinBoxXZoom.setValue(
            int(
                graphic_view.sceneRect().width()
                / graphic_view.view_rect().width()
                * 100
            )
        )
        self.ui.spinBoxXZoom.blockSignals(False)

    @pyqtSlot()
    def on_signal_zoomed(self):
        self.__handle_graphic_view_zoomed(self.ui.gvSignal)

    @pyqtSlot()
    def on_spectrum_zoomed(self):
        self.__handle_graphic_view_zoomed(self.ui.gvSpectrogram)

    @pyqtSlot(int)
    def on_spinbox_x_zoom_value_changed(self, value: int):
        graphic_view = (
            self.ui.gvSpectrogram if self.spectrogram_is_active else self.ui.gvSignal
        )
        zoom_factor = value / 100
        current_factor = (
            graphic_view.sceneRect().width() / graphic_view.view_rect().width()
        )
        graphic_view.zoom(zoom_factor / current_factor)

    @pyqtSlot()
    def on_btn_close_signal_clicked(self):
        self.my_close()

    @pyqtSlot()
    def on_set_noise_in_graphic_view_clicked(self):
        self.setCursor(Qt.WaitCursor)
        start = self.ui.gvSignal.selection_area.x
        end = start + self.ui.gvSignal.selection_area.width

        new_thresh = self.signal.calc_relative_noise_threshold_from_range(start, end)
        self.ui.spinBoxNoiseTreshold.setValue(new_thresh)
        self.ui.spinBoxNoiseTreshold.editingFinished.emit()
        self.unsetCursor()

    @pyqtSlot(float)
    def on_signal_center_spacing_changed(self, value: float):
        self.ui.spinBoxCenterSpacing.setValue(value)
        if self.ui.gvSignal.scene_type == 1:
            self.ui.gvSignal.scene().redraw_legend()

    @pyqtSlot()
    def on_noise_threshold_changed(self):
        self.ui.spinBoxNoiseTreshold.setValue(self.signal.noise_threshold_relative)
        minimum = self.signal.noise_min_plot
        maximum = self.signal.noise_max_plot
        if (
            self.ui.cbSignalView.currentIndex() == 0
            or self.ui.cbSignalView.currentIndex() == 3
        ):
            # Draw Noise only in analog and I/Q view
            self.ui.gvSignal.scene().draw_noise_area(minimum, maximum - minimum)

    @pyqtSlot(int)
    def on_spinbox_selection_start_value_changed(self, value: int):
        if self.spectrogram_is_active:
            self.ui.gvSpectrogram.set_vertical_selection(
                y=self.ui.gvSpectrogram.sceneRect().height() - value
            )
            self.ui.gvSpectrogram.emit_selection_size_changed()
            self.ui.gvSpectrogram.selection_area.finished = True
        else:
            self.ui.gvSignal.set_horizontal_selection(x=value)
            self.ui.gvSignal.selection_area.finished = True
            self.ui.gvSignal.emit_selection_size_changed()

    @pyqtSlot(int)
    def on_spinbox_selection_end_value_changed(self, value: int):
        if self.spectrogram_is_active:
            self.ui.gvSpectrogram.set_vertical_selection(
                h=self.ui.spinBoxSelectionStart.value() - value
            )
            self.ui.gvSpectrogram.emit_selection_size_changed()
            self.ui.gvSpectrogram.selection_area.finished = True
        else:
            self.ui.gvSignal.set_horizontal_selection(
                w=value - self.ui.spinBoxSelectionStart.value()
            )
            self.ui.gvSignal.selection_area.finished = True
            self.ui.gvSignal.emit_selection_size_changed()

    @pyqtSlot()
    def on_protocol_updated(self):
        self.ui.gvSignal.redraw_view()  # Participants may have changed
        self.ui.txtEdProto.setEnabled(True)
        self.ui.txtEdProto.setHtml(self.proto_analyzer.plain_to_html(self.proto_view))

    @pyqtSlot()
    def handle_protocol_sync_changed(self):
        self.sync_protocol = self.ui.chkBoxSyncSelection.isChecked()

    @pyqtSlot()
    def set_protocol_visibility(self):
        checked = self.ui.chkBoxShowProtocol.isChecked()

        if checked:
            self.show_protocol()
            self.ui.cbProtoView.setEnabled(True)
            # Disabled because never used
            # self.ui.chkBoxSyncSelection.show()
            self.ui.txtEdProto.show()
        else:
            self.ui.txtEdProto.hide()
            self.ui.chkBoxSyncSelection.hide()
            self.ui.cbProtoView.setEnabled(False)

        qApp.processEvents()
        self.on_slider_y_scale_value_changed()

    @pyqtSlot()
    def on_cb_signal_view_index_changed(self):
        self.setCursor(Qt.WaitCursor)

        self.__set_spectrogram_adjust_widgets_visibility()

        if self.ui.cbSignalView.currentText().lower() == "spectrogram":
            self.ui.stackedWidget.setCurrentWidget(self.ui.pageSpectrogram)
            self.draw_spectrogram(show_full_scene=True)
            self.__set_selected_bandwidth()
            self.ui.labelRSSI.hide()
        else:
            self.ui.stackedWidget.setCurrentWidget(self.ui.pageSignal)
            self.ui.gvSignal.scene_type = self.ui.cbSignalView.currentIndex()
            self.scene_manager.mod_type = self.signal.modulation_type
            self.ui.gvSignal.redraw_view(reinitialize=True)
            self.ui.labelRSSI.show()

            self.ui.gvSignal.auto_fit_view()
            self.ui.gvSignal.refresh_selection_area()
            qApp.processEvents()
            self.on_slider_y_scale_value_changed()  # apply YScale to new view
            self.__set_samples_in_view()
            self.__set_duration()

        self.unsetCursor()

    @pyqtSlot()
    def on_btn_autodetect_clicked(self):
        self.ui.btnAutoDetect.setEnabled(False)
        self.setCursor(Qt.WaitCursor)

        try:
            detect_modulation = self.detect_modulation_action.isChecked()
        except AttributeError:
            detect_modulation = False

        try:
            detect_noise = self.detect_noise_action.isChecked()
        except AttributeError:
            detect_noise = False
        success = self.signal.auto_detect(
            detect_modulation=detect_modulation, detect_noise=detect_noise
        )

        self.ui.btnAutoDetect.setEnabled(True)
        self.unsetCursor()
        if not success:
            Errors.generic_error(
                self.tr("Autodetection failed"),
                self.tr("Failed to autodetect parameters for this signal."),
            )

    @pyqtSlot()
    def on_btn_replay_clicked(self):
        project_manager = self.project_manager
        try:
            dialog = SendDialog(
                project_manager, modulated_data=self.signal.iq_array, parent=self
            )
        except OSError as e:
            logger.error(repr(e))
            return

        if dialog.has_empty_device_list:
            Errors.no_device()
            dialog.close()
            return

        dialog.device_parameters_changed.connect(project_manager.set_device_parameters)
        dialog.show()
        dialog.graphics_view.show_full_scene(reinitialize=True)

    @pyqtSlot(int, int)
    def update_selection_area(self, start, end):
        self.update_number_selected_samples()
        self.ui.spinBoxSelectionStart.blockSignals(True)
        self.ui.spinBoxSelectionStart.setValue(start)
        self.ui.spinBoxSelectionStart.blockSignals(False)
        self.ui.spinBoxSelectionEnd.blockSignals(True)
        self.ui.spinBoxSelectionEnd.setValue(end)
        self.ui.spinBoxSelectionEnd.blockSignals(False)

    @pyqtSlot()
    def refresh_protocol(self):
        self.show_protocol(refresh=True)

    @pyqtSlot(int)
    def on_combo_box_proto_view_index_changed(self, index: int):
        old_view = self.ui.txtEdProto.cur_view
        self.ui.txtEdProto.cur_view = index
        self.show_protocol(old_view=old_view)

    @pyqtSlot(float)
    def set_center(self, th):
        self.ui.spinBoxCenterOffset.setValue(th)
        self.ui.spinBoxCenterOffset.editingFinished.emit()

    def set_roi_from_protocol_analysis(
        self, start_message, start_pos, end_message, end_pos, view_type
    ):
        if not self.proto_analyzer:
            return

        if not self.ui.chkBoxShowProtocol.isChecked():
            self.ui.chkBoxShowProtocol.setChecked(True)
            self.set_protocol_visibility()

        self.ui.cbProtoView.setCurrentIndex(view_type)

        if view_type == 1:
            # Hex View
            start_pos *= 4
            end_pos *= 4
        elif view_type == 2:
            # ASCII View
            start_pos *= 8
            end_pos *= 8

        sample_pos, num_samples = self.proto_analyzer.get_samplepos_of_bitseq(
            start_message, start_pos, end_message, end_pos, True
        )
        self.protocol_selection_is_updateable = False
        if sample_pos != -1:
            if self.jump_sync and self.sync_protocol:
                self.ui.gvSignal.centerOn(sample_pos, self.ui.gvSignal.y_center)
                self.ui.gvSignal.set_horizontal_selection(sample_pos, num_samples)
                self.ui.gvSignal.centerOn(
                    sample_pos + num_samples, self.ui.gvSignal.y_center
                )
            else:
                self.ui.gvSignal.set_horizontal_selection(sample_pos, num_samples)

            self.ui.gvSignal.zoom_to_selection(sample_pos, sample_pos + num_samples)
        else:
            self.ui.gvSignal.clear_horizontal_selection()

        self.protocol_selection_is_updateable = True
        self.update_protocol_selection_from_roi()

    @pyqtSlot()
    def update_roi_from_protocol_selection(self):
        text_edit = self.ui.txtEdProto
        start_pos, end_pos = (
            text_edit.textCursor().selectionStart(),
            text_edit.textCursor().selectionEnd(),
        )
        if start_pos == end_pos == -1:
            return

        forward_selection = (
            text_edit.textCursor().anchor() <= text_edit.textCursor().position()
        )

        if start_pos > end_pos:
            start_pos, end_pos = end_pos, start_pos

        text = text_edit.toPlainText()

        start_message = text[:start_pos].count("\n")
        end_message = start_message + text[start_pos:end_pos].count("\n")
        newline_pos = text[:start_pos].rfind("\n")

        if newline_pos != -1:
            start_pos -= newline_pos + 1

        newline_pos = text[:end_pos].rfind("\n")
        if newline_pos != -1:
            end_pos -= newline_pos + 1

        factor = 1 if text_edit.cur_view == 0 else 4 if text_edit.cur_view == 1 else 8
        start_pos *= factor
        end_pos *= factor

        try:
            include_last_pause = False
            s = text_edit.textCursor().selectionStart()
            e = text_edit.textCursor().selectionEnd()
            if s > e:
                s, e = e, s

            selected_text = text[s:e]

            last_newline = selected_text.rfind("\n")
            if last_newline == -1:
                last_newline = 0

            if selected_text.endswith(" "):
                end_pos -= 1
            elif selected_text.endswith(" \t"):
                end_pos -= 2

            if "[" in selected_text[last_newline:]:
                include_last_pause = True

            sample_pos, num_samples = self.proto_analyzer.get_samplepos_of_bitseq(
                start_message, start_pos, end_message, end_pos, include_last_pause
            )

        except IndexError:
            return

        self.ui.gvSignal.blockSignals(True)
        if sample_pos != -1:
            if self.jump_sync and self.sync_protocol:
                self.ui.gvSignal.centerOn(sample_pos, self.ui.gvSignal.y_center)
                self.ui.gvSignal.set_horizontal_selection(sample_pos, num_samples)
                if (
                    forward_selection
                ):  # Forward Selection --> Center ROI to End of Selection
                    self.ui.gvSignal.centerOn(
                        sample_pos + num_samples, self.ui.gvSignal.y_center
                    )
                else:  # Backward Selection --> Center ROI to Start of Selection
                    self.ui.gvSignal.centerOn(sample_pos, self.ui.gvSignal.y_center)
            else:
                self.ui.gvSignal.set_horizontal_selection(sample_pos, num_samples)
        else:
            self.ui.gvSignal.clear_horizontal_selection()
        self.ui.gvSignal.blockSignals(False)

        self.update_number_selected_samples()

    def zoom_to_roi(self):
        roi = self.ui.gvSignal.selection_area
        start, end = roi.x, roi.x + roi.width
        self.ui.gvSignal.zoom_to_selection(start, end)

    @pyqtSlot()
    def start_proto_selection_timer(self):
        self.proto_selection_timer.start()

    @pyqtSlot()
    def update_protocol_selection_from_roi(self):
        protocol = self.proto_analyzer

        if (
            protocol is None
            or protocol.messages is None
            or not self.ui.chkBoxShowProtocol.isChecked()
        ):
            return

        start = self.ui.gvSignal.selection_area.x
        w = self.ui.gvSignal.selection_area.width

        if w < 0:
            start += w
            w = -w

        c = self.ui.txtEdProto.textCursor()
        self.jump_sync = False
        self.ui.txtEdProto.blockSignals(True)

        try:
            (
                start_message,
                start_index,
                end_message,
                end_index,
            ) = protocol.get_bitseq_from_selection(start, w)
        except IndexError:
            c.clearSelection()
            self.ui.txtEdProto.setTextCursor(c)
            self.jump_sync = True
            self.ui.txtEdProto.blockSignals(False)
            return

        if (
            start_message == -1
            or end_index == -1
            or start_index == -1
            or end_message == -1
        ):
            c.clearSelection()
            self.ui.txtEdProto.setTextCursor(c)
            self.jump_sync = True
            self.ui.txtEdProto.blockSignals(False)
            return

        start_index = int(
            protocol.convert_index(start_index, 0, self.proto_view, True)[0]
        )
        end_index = (
            int(
                math.ceil(
                    protocol.convert_index(end_index, 0, self.proto_view, True)[1]
                )
            )
            + 1
        )
        text = self.ui.txtEdProto.toPlainText()
        n = 0
        message_pos = 0
        c.setPosition(0)

        for i, t in enumerate(text):
            message_pos += 1
            if t == "\n":
                n += 1
                message_pos = 0

            if n == start_message and message_pos == start_index:
                c.setPosition(i + 1, QTextCursor.MoveAnchor)

            if n == end_message and message_pos == end_index:
                c.setPosition(i, QTextCursor.KeepAnchor)
                break

        self.ui.txtEdProto.setTextCursor(c)
        self.ui.txtEdProto.blockSignals(False)
        self.jump_sync = True

    def __set_samples_in_view(self):
        if self.spectrogram_is_active:
            self.ui.lSamplesInView.setText(
                "{0:n}".format(int(self.ui.gvSpectrogram.view_rect().width()))
            )
            self.ui.lSamplesTotal.setText(
                "{0:n}".format(self.ui.gvSpectrogram.width_spectrogram)
            )
        else:
            self.ui.lSamplesInView.setText(
                "{0:n}".format(int(self.ui.gvSignal.view_rect().width()))
            )
            self.ui.lSamplesTotal.setText("{0:n}".format(self.signal.num_samples))

    def refresh_signal(self, draw_full_signal=False):
        self.draw_signal(draw_full_signal)

        self.__set_samples_in_view()
        self.update_number_selected_samples()
        self.on_slider_y_scale_value_changed()

    @pyqtSlot(float)
    def on_signal_center_changed(self, center):
        self.ui.gvSignal.y_sep = -center

        if self.ui.cbSignalView.currentIndex() > 0:
            self.scene_manager.scene.draw_sep_area(-self.signal.center_thresholds)
        self.ui.spinBoxCenterOffset.blockSignals(False)
        self.ui.spinBoxCenterOffset.setValue(center)

    def on_spinbox_noise_threshold_editing_finished(self):
        if (
            self.signal is not None
            and self.signal.noise_threshold_relative
            != self.ui.spinBoxNoiseTreshold.value()
        ):
            noise_action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="noise_threshold_relative",
                parameter_value=self.ui.spinBoxNoiseTreshold.value(),
            )
            self.undo_stack.push(noise_action)

    def contextMenuEvent(self, event: QContextMenuEvent):
        if self.signal is None:
            return

        menu = QMenu()
        apply_to_all_action = menu.addAction(
            self.tr("Apply values (BitLen, 0/1-Threshold, Tolerance) to all signals")
        )
        menu.addSeparator()
        auto_detect_action = menu.addAction(self.tr("Auto-Detect signal parameters"))
        action = menu.exec_(self.mapToGlobal(event.pos()))
        if action == apply_to_all_action:
            self.setCursor(Qt.WaitCursor)
            self.apply_to_all_clicked.emit(self.signal)
            self.unsetCursor()
        elif action == auto_detect_action:
            self.setCursor(Qt.WaitCursor)
            self.signal.auto_detect(detect_modulation=False, detect_noise=False)
            self.unsetCursor()

    def show_modulation_type(self):
        self.ui.cbModulationType.blockSignals(True)
        self.ui.cbModulationType.setCurrentText(self.signal.modulation_type)
        self.ui.cbModulationType.blockSignals(False)

    def on_participant_changed(self):
        if hasattr(self, "proto_analyzer") and self.proto_analyzer:
            self.proto_analyzer.qt_signals.protocol_updated.emit()

    def resizeEvent(self, event: QResizeEvent):
        old_width, new_width = max(1, event.oldSize().width()), max(
            1, event.size().width()
        )
        super().resizeEvent(event)
        self.on_slider_y_scale_value_changed()

        # Force update of GVS, when size changed e.g. when Project Tree is opened
        if not self.spectrogram_is_active:
            self.ui.gvSignal.zoom(new_width / old_width, zoom_to_mouse_cursor=False)

    def set_center_spacing_visibility(self):
        visible = self.ui.spinBoxBitsPerSymbol.value() > 1
        self.ui.spinBoxCenterSpacing.setVisible(visible)
        self.ui.lCenterSpacing.setVisible(visible)

    @pyqtSlot()
    def on_info_btn_clicked(self):
        sdc = SignalDetailsDialog(self.signal, self)
        sdc.show()

    @pyqtSlot(str)
    def on_combobox_modulation_type_text_changed(self, txt: str):
        if txt != self.signal.modulation_type:
            modulation_action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="modulation_type",
                parameter_value=txt,
            )

            self.undo_stack.push(modulation_action)

            self.scene_manager.mod_type = txt
            if self.ui.cbSignalView.currentIndex() == 1:
                self.scene_manager.init_scene()
                self.on_slider_y_scale_value_changed()

        self.ui.btnAdvancedModulationSettings.setVisible(
            self.ui.cbModulationType.currentText() in ("ASK", "PSK")
        )

    @pyqtSlot()
    def on_signal_data_changed_before_save(self):
        font = self.ui.lineEditSignalName.font()
        self.ui.gvSignal.auto_fit_on_resize_is_blocked = True

        if self.signal.changed:
            font.setBold(True)
            self.ui.btnSaveSignal.show()
        else:
            font.setBold(False)
            self.ui.btnSaveSignal.hide()
            for i in range(self.undo_stack.count()):
                cmd = self.undo_stack.command(i)
                if isinstance(cmd, EditSignalAction):
                    # https://github.com/jopohl/urh/issues/570
                    cmd.signal_was_changed = True

        qApp.processEvents()
        self.ui.gvSignal.auto_fit_on_resize_is_blocked = False

        self.ui.lineEditSignalName.setFont(font)

    @pyqtSlot()
    def on_btn_show_hide_start_end_clicked(self):
        show = self.ui.btnShowHideStartEnd.isChecked()
        if show:
            self.ui.btnShowHideStartEnd.setIcon(QIcon.fromTheme("arrow-down-double"))
            self.ui.verticalLayout.insertItem(2, self.ui.additionalInfos)
        else:
            self.ui.btnShowHideStartEnd.setIcon(QIcon.fromTheme("arrow-up-double"))
            self.ui.verticalLayout.removeItem(self.ui.additionalInfos)

        for i in range(self.ui.additionalInfos.count()):
            try:
                self.ui.additionalInfos.itemAt(i).widget().setVisible(show)
            except AttributeError:
                pass

    @pyqtSlot()
    def on_spinbox_tolerance_editing_finished(self):
        if self.signal.tolerance != self.ui.spinBoxTolerance.value():
            self.ui.spinBoxTolerance.blockSignals(True)
            tolerance_action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="tolerance",
                parameter_value=self.ui.spinBoxTolerance.value(),
            )
            self.undo_stack.push(tolerance_action)
            self.ui.spinBoxTolerance.blockSignals(False)

    @pyqtSlot()
    def on_spinbox_samples_per_symbol_editing_finished(self):
        if self.signal.samples_per_symbol != self.ui.spinBoxSamplesPerSymbol.value():
            self.ui.spinBoxSamplesPerSymbol.blockSignals(True)
            action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="samples_per_symbol",
                parameter_value=self.ui.spinBoxSamplesPerSymbol.value(),
            )
            self.undo_stack.push(action)
            self.ui.spinBoxSamplesPerSymbol.blockSignals(False)

    @pyqtSlot()
    def on_spinbox_bits_per_symbol_editing_finished(self):
        if self.signal.bits_per_symbol != self.ui.spinBoxBitsPerSymbol.value():
            self.ui.spinBoxBitsPerSymbol.blockSignals(True)
            bits_per_symbol_action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="bits_per_symbol",
                parameter_value=self.ui.spinBoxBitsPerSymbol.value(),
            )
            self.undo_stack.push(bits_per_symbol_action)
            self.ui.spinBoxBitsPerSymbol.blockSignals(False)

            if self.ui.gvSignal.scene_type == 1:
                self.ui.gvSignal.scene().draw_sep_area(-self.signal.center_thresholds)

            self.set_center_spacing_visibility()

    @pyqtSlot()
    def on_spinbox_center_editing_finished(self):
        if self.signal.center != self.ui.spinBoxCenterOffset.value():
            self.ui.spinBoxCenterOffset.blockSignals(True)
            center_action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="center",
                parameter_value=self.ui.spinBoxCenterOffset.value(),
            )
            self.undo_stack.push(center_action)
            self.ui.spinBoxCenterOffset.blockSignals(False)

    @pyqtSlot()
    def on_spinbox_spacing_value_changed(self):
        if self.ui.gvSignal.scene_type == 1:
            thresholds = self.signal.get_thresholds_for_center(
                self.signal.center, self.ui.spinBoxCenterSpacing.value()
            )
            self.ui.gvSignal.scene().draw_sep_area(-thresholds)

    @pyqtSlot()
    def on_spinbox_spacing_editing_finished(self):
        if self.signal.center_spacing != self.ui.spinBoxCenterSpacing.value():
            self.ui.spinBoxCenterSpacing.blockSignals(True)
            center_spacing_action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="center_spacing",
                parameter_value=self.ui.spinBoxCenterSpacing.value(),
            )
            self.undo_stack.push(center_spacing_action)
            self.ui.spinBoxCenterSpacing.blockSignals(False)

            if self.ui.gvSignal.scene_type == 1:
                self.ui.gvSignal.scene().draw_sep_area(-self.signal.center_thresholds)

    @pyqtSlot()
    def refresh(self, draw_full_signal=False):
        self.refresh_signal(draw_full_signal=draw_full_signal)
        self.refresh_signal_information(block=True)
        self.show_protocol(refresh=True)

    @pyqtSlot()
    def on_btn_filter_clicked(self):
        if self.apply_filter_to_selection_only.isChecked():
            start, end = (
                self.ui.gvSignal.selection_area.start,
                self.ui.gvSignal.selection_area.end,
            )
        else:
            start, end = 0, self.signal.num_samples

        filter_action = EditSignalAction(
            signal=self.signal,
            mode=EditAction.filter,
            start=start,
            end=end,
            dsp_filter=self.dsp_filter,
            protocol=self.proto_analyzer,
        )
        self.undo_stack.push(filter_action)

    @pyqtSlot()
    def on_configure_filter_action_triggered(self):
        self.filter_dialog.set_dsp_filter_status(self.dsp_filter.filter_type)
        self.filter_dialog.exec()

    @pyqtSlot(Filter)
    def on_filter_dialog_filter_accepted(self, dsp_filter: Filter):
        if dsp_filter is not None:
            self.dsp_filter = dsp_filter
            self.set_filter_button_caption()

    @pyqtSlot()
    def on_spectrogram_update_timer_timeout(self):
        self.draw_spectrogram(show_full_scene=True)

    @pyqtSlot(float)
    def on_gv_spectrogram_y_scale_changed(self, scale: float):
        self.ui.sliderYScale.blockSignals(True)
        self.ui.sliderYScale.setValue(self.ui.sliderYScale.value() * scale)
        self.ui.sliderYScale.blockSignals(False)

    @pyqtSlot(float, float)
    def on_bandpass_filter_triggered(self, f_low: float, f_high: float):
        self.filter_abort_wanted = False

        QApplication.instance().setOverrideCursor(Qt.WaitCursor)
        filter_bw = Filter.read_configured_filter_bw()
        filtered = Array("f", 2 * self.signal.num_samples)
        p = Process(
            target=perform_filter,
            args=(
                filtered,
                self.signal.iq_array.as_complex64(),
                f_low,
                f_high,
                filter_bw,
            ),
        )
        p.daemon = True
        p.start()

        while p.is_alive():
            QApplication.instance().processEvents()

            if self.filter_abort_wanted:
                p.terminate()
                p.join()
                QApplication.instance().restoreOverrideCursor()
                return

            time.sleep(0.1)

        filtered = np.frombuffer(filtered.get_obj(), dtype=np.complex64)
        signal = self.signal.create_new(
            new_data=filtered.astype(np.complex64), new_timestamp=self.signal.timestamp
        )
        signal.name = (
            self.signal.name
            + " filtered with f_low={0:.4n} f_high={1:.4n} bw={2:.4n}".format(
                f_low, f_high, filter_bw
            )
        )
        self.signal_created.emit(signal)
        QApplication.instance().restoreOverrideCursor()

    def on_signal_data_edited(self):
        self.refresh_signal()
        self.ui.gvSpectrogram.scene_manager.samples_need_update = True

    @pyqtSlot()
    def on_signal_sample_rate_changed(self):
        if self.spectrogram_is_active:
            self.__set_selected_bandwidth()
        else:
            self.__set_duration()

        self.show_protocol()  # update times

    @pyqtSlot(int)
    def on_pause_threshold_edited(self, pause_threshold: int):
        if self.signal.pause_threshold != pause_threshold:
            pause_threshold_action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="pause_threshold",
                parameter_value=pause_threshold,
            )
            self.undo_stack.push(pause_threshold_action)

    @pyqtSlot(int)
    def on_message_length_divisor_edited(self, message_length_divisor: int):
        if self.signal.message_length_divisor != message_length_divisor:
            message_length_divisor_action = ChangeSignalParameter(
                signal=self.signal,
                protocol=self.proto_analyzer,
                parameter_name="message_length_divisor",
                parameter_value=message_length_divisor,
            )
            self.undo_stack.push(message_length_divisor_action)

    def get_advanced_modulation_settings_dialog(self):
        dialog = AdvancedModulationOptionsDialog(
            self.signal.pause_threshold, self.signal.message_length_divisor, parent=self
        )
        dialog.pause_threshold_edited.connect(self.on_pause_threshold_edited)
        dialog.message_length_divisor_edited.connect(
            self.on_message_length_divisor_edited
        )
        return dialog

    def get_costas_dialog(self):
        dialog = CostaOptionsDialog(self.signal.costas_loop_bandwidth, parent=self)
        dialog.accepted.connect(self.on_costas_dialog_accepted)
        return dialog

    @pyqtSlot()
    def on_costas_dialog_accepted(self):
        sender = self.sender()
        assert isinstance(sender, CostaOptionsDialog)
        self.signal.costas_loop_bandwidth = sender.costas_loop_bandwidth

    @pyqtSlot()
    def on_btn_advanced_modulation_settings_clicked(self):
        if self.ui.cbModulationType.currentText() == "ASK":
            dialog = self.get_advanced_modulation_settings_dialog()
        elif self.ui.cbModulationType.currentText() == "PSK":
            dialog = self.get_costas_dialog()
        else:
            raise ValueError("No additional settings available")
        dialog.exec_()

    @pyqtSlot()
    def on_export_fta_wanted(self):
        try:
            initial_name = self.signal.name + "-spectrogram.ft"
        except Exception as e:
            logger.exception(e)
            initial_name = "spectrogram.ft"

        filename = FileOperator.ask_save_file_name(
            initial_name, caption="Export spectrogram"
        )
        if not filename:
            return
        QApplication.setOverrideCursor(Qt.WaitCursor)
        try:
            self.ui.gvSpectrogram.scene_manager.spectrogram.export_to_fta(
                sample_rate=self.signal.sample_rate,
                filename=filename,
                include_amplitude=filename.endswith(".fta"),
            )
        except Exception as e:
            Errors.exception(e)
        finally:
            QApplication.restoreOverrideCursor()



from PyQt5.QtCore import QModelIndex, QAbstractItemModel, Qt, pyqtSlot
from PyQt5.QtWidgets import (
    QStyledItemDelegate,
    QWidget,
    QStyleOptionViewItem,
    QCheckBox,
)


class CheckBoxDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.enabled = True

    def createEditor(
        self, parent: QWidget, option: QStyleOptionViewItem, index: QModelIndex
    ):
        editor = QCheckBox(parent)
        editor.stateChanged.connect(self.stateChanged)
        return editor

    def setEditorData(self, editor: QCheckBox, index: QModelIndex):
        editor.blockSignals(True)
        editor.setChecked(index.model().data(index))
        self.enabled = editor.isChecked()
        editor.blockSignals(False)

    def setModelData(
        self, editor: QCheckBox, model: QAbstractItemModel, index: QModelIndex
    ):
        model.setData(index, editor.isChecked(), Qt.EditRole)

    @pyqtSlot()
    def stateChanged(self):
        self.commitData.emit(self.sender())



import os
import time
from datetime import datetime
from threading import Thread

import numpy as np
from PyQt5.QtCore import pyqtSignal, QObject
from urh.cythonext.signal_functions import grab_pulse_lens

from urh.ainterpretation import AutoInterpretation
from urh.dev.BackendHandler import BackendHandler, Backends
from urh.dev.VirtualDevice import VirtualDevice, Mode
from urh.signalprocessing.IQArray import IQArray
from urh.signalprocessing.Message import Message
from urh.signalprocessing.ProtocolAnalyzer import ProtocolAnalyzer
from urh.signalprocessing.Signal import Signal
from urh.util.Logger import logger


class ProtocolSniffer(ProtocolAnalyzer, QObject):
    """
    This class is used for live sniffing a protocol
    with certain signal parameters.
    """

    started = pyqtSignal()
    stopped = pyqtSignal()
    message_sniffed = pyqtSignal(int)

    BUFFER_SIZE_MB = 100

    def __init__(
        self,
        samples_per_symbol: int,
        center: float,
        center_spacing: float,
        noise: float,
        tolerance: int,
        modulation_type: str,
        bits_per_symbol: int,
        device: str,
        backend_handler: BackendHandler,
        network_raw_mode=False,
    ):
        signal = Signal("", "LiveSignal")
        signal.samples_per_symbol = samples_per_symbol
        signal.center = center
        signal.center_spacing = center_spacing
        signal.noise_threshold = noise
        signal.tolerance = tolerance
        signal.silent_set_modulation_type(modulation_type)
        signal.bits_per_symbol = bits_per_symbol
        ProtocolAnalyzer.__init__(self, signal)
        QObject.__init__(self, None)

        self.network_raw_mode = network_raw_mode
        self.backend_handler = backend_handler
        self.rcv_device = VirtualDevice(
            self.backend_handler,
            device,
            Mode.receive,
            resume_on_full_receive_buffer=True,
            raw_mode=network_raw_mode,
        )

        signal.iq_array = IQArray(None, self.rcv_device.data_type, 0)

        self.sniff_thread = Thread(target=self.check_for_data, daemon=True)

        self.rcv_device.started.connect(self.__emit_started)
        self.rcv_device.stopped.connect(self.__emit_stopped)

        self.__buffer = IQArray(None, np.float32, 0)
        self.__init_buffer()
        self.__current_buffer_index = 0

        self.reading_data = False
        self.adaptive_noise = False
        self.automatic_center = False

        self.pause_length = 0
        self.is_running = False

        self.store_messages = True

        self.__sniff_file = ""
        self.__store_data = True

    def __add_to_buffer(self, data: np.ndarray):
        n = len(data)
        if n + self.__current_buffer_index > len(self.__buffer):
            n = len(self.__buffer) - self.__current_buffer_index - 1
            logger.warning("Buffer of protocol sniffer is full")

        self.__buffer[
            self.__current_buffer_index : self.__current_buffer_index + n
        ] = data[:n]
        self.__current_buffer_index += n

    def __clear_buffer(self):
        self.__current_buffer_index = 0

    def __buffer_is_full(self):
        return self.__current_buffer_index >= len(self.__buffer) - 2

    def __init_buffer(self):
        self.__buffer = IQArray(
            None, self.rcv_device.data_type, int(self.BUFFER_SIZE_MB * 1000 * 1000 / 8)
        )
        self.__current_buffer_index = 0

    def decoded_to_string(self, view: int, start=0, include_timestamps=True):
        result = []
        for msg in self.messages[start:]:
            result.append(self.message_to_string(msg, view, include_timestamps))
        return "\n".join(result)

    def message_to_string(self, message: Message, view: int, include_timestamps=True):
        msg_str_data = []
        if include_timestamps:
            msg_date = datetime.fromtimestamp(message.timestamp)
            msg_str_data.append(msg_date.strftime("[%Y-%m-%d %H:%M:%S.%f]"))
        msg_str_data.append(
            message.view_to_string(view, decoded=True, show_pauses=False)
        )
        return " ".join(msg_str_data)

    @property
    def sniff_file(self):
        return self.__sniff_file

    @sniff_file.setter
    def sniff_file(self, val):
        self.__sniff_file = val
        if self.__sniff_file:
            self.__store_data = False

    @property
    def device_name(self):
        return self.rcv_device.name

    @device_name.setter
    def device_name(self, value: str):
        if value != self.rcv_device.name:
            self.rcv_device.free_data()
            self.rcv_device = VirtualDevice(
                self.backend_handler,
                value,
                Mode.receive,
                device_ip="192.168.10.2",
                resume_on_full_receive_buffer=True,
                raw_mode=self.network_raw_mode,
            )
            self.rcv_device.started.connect(self.__emit_started)
            self.rcv_device.stopped.connect(self.__emit_stopped)

            self.signal.iq_array = IQArray(None, self.rcv_device.data_type, 0)

            self.__init_buffer()

    def sniff(self):
        self.is_running = True
        self.rcv_device.start()
        self.sniff_thread = Thread(target=self.check_for_data, daemon=True)
        self.sniff_thread.start()

    def check_for_data(self):
        old_index = 0
        while self.is_running:
            time.sleep(0.01)
            if self.rcv_device.is_raw_mode:
                if old_index <= self.rcv_device.current_index:
                    data = self.rcv_device.data[
                        old_index : self.rcv_device.current_index
                    ]
                else:
                    data = np.concatenate(
                        (
                            self.rcv_device.data[old_index:],
                            self.rcv_device.data[: self.rcv_device.current_index],
                        )
                    )
                old_index = self.rcv_device.current_index
                self.__demodulate_data(data)
            elif self.rcv_device.backend == Backends.network:
                # We receive the bits here
                for bit_str in self.rcv_device.data:
                    msg = Message.from_plain_bits_str(bit_str)
                    msg.decoder = self.decoder
                    self.messages.append(msg)
                    self.message_sniffed.emit(len(self.messages) - 1)

                self.rcv_device.free_data()  # do not store received bits twice

            if self.sniff_file and not os.path.isdir(self.sniff_file):
                plain_bits_str = self.plain_bits_str
                if plain_bits_str:
                    with open(self.sniff_file, "a") as f:
                        f.write("\n".join(plain_bits_str) + "\n")

            if not self.__store_data:
                self.messages.clear()

    def __demodulate_data(self, data):
        """
        Demodulates received IQ data and adds demodulated bits to messages
        :param data:
        :return:
        """
        if len(data) == 0:
            return

        power_spectrum = data.real**2.0 + data.imag**2.0
        is_above_noise = np.sqrt(np.mean(power_spectrum)) > self.signal.noise_threshold

        if self.adaptive_noise and not is_above_noise:
            self.signal.noise_threshold = (
                0.9 * self.signal.noise_threshold
                + 0.1 * np.sqrt(np.max(power_spectrum))
            )

        if is_above_noise:
            self.__add_to_buffer(data)
            self.pause_length = 0
            if not self.__buffer_is_full():
                return
        else:
            self.pause_length += len(data)
            if self.pause_length < 10 * self.signal.samples_per_symbol:
                self.__add_to_buffer(data)
                if not self.__buffer_is_full():
                    return

        if self.__current_buffer_index == 0:
            return

        # clear cache and start a new message
        self.signal.iq_array = IQArray(self.__buffer[0 : self.__current_buffer_index])
        self.signal.timestamp = time.time() - (
            len(self.signal.iq_array) / self.rcv_device.sample_rate
        )  # timestamp of the first sample in our buffer
        self.__clear_buffer()
        self.signal._qad = None

        samples_per_symbol = self.signal.samples_per_symbol
        if self.automatic_center:
            self.signal.center = AutoInterpretation.detect_center(
                self.signal.qad, max_size=150 * samples_per_symbol
            )

        ppseq = grab_pulse_lens(
            self.signal.qad,
            self.signal.center,
            self.signal.tolerance,
            self.signal.modulation_type,
            self.signal.samples_per_symbol,
            self.signal.bits_per_symbol,
            self.signal.center_spacing,
        )

        bit_data, pauses, bit_sample_pos = self._ppseq_to_bits(
            ppseq,
            samples_per_symbol,
            self.signal.bits_per_symbol,
            write_bit_sample_pos=True,
        )

        i = 0
        for bits, pause in zip(bit_data, pauses):
            message_timestamp = self.signal.timestamp + (
                bit_sample_pos[i][0] / self.rcv_device.sample_rate
            )
            message = Message(
                bits,
                pause,
                samples_per_symbol=samples_per_symbol,
                message_type=self.default_message_type,
                decoder=self.decoder,
                timestamp=message_timestamp,
            )
            self.messages.append(message)
            self.message_sniffed.emit(len(self.messages) - 1)
            i += 1

    def stop(self):
        self.is_running = False
        self.rcv_device.stop("Stopping receiving due to user interaction")
        if self.sniff_thread.is_alive():
            self.sniff_thread.join(0.1)
        if self.sniff_thread.is_alive():
            logger.error("Sniff thread is still alive")

    def clear(self):
        self.__clear_buffer()
        self.messages.clear()

    def __emit_started(self):
        self.started.emit()

    def __emit_stopped(self):
        if hasattr(self, "stopped"):
            self.stopped.emit()



from PyQt5.QtCore import pyqtSlot, Qt, QItemSelection, QItemSelectionModel
from PyQt5.QtGui import QKeySequence, QIcon, QContextMenuEvent
from PyQt5.QtWidgets import QTableView, QAction, QMenu

from urh import settings
from urh.models.ParticipantTableModel import ParticipantTableModel
from urh.ui.delegates.ComboBoxDelegate import ComboBoxDelegate


class ParticipantTableView(QTableView):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.remove_action = QAction("Remove selected participants", self)
        self.remove_action.setShortcut(QKeySequence.Delete)
        self.remove_action.setIcon(QIcon.fromTheme("list-remove"))
        self.remove_action.setShortcutContext(Qt.WidgetWithChildrenShortcut)
        self.remove_action.triggered.connect(self.on_remove_action_triggered)
        self.addAction(self.remove_action)

    @property
    def selected_columns(self) -> (int, int):
        selection = self.selectionModel().selection()
        if selection.isEmpty():
            return 0, self.model().columnCount() - 1

        return min([rng.left() for rng in selection]), max(
            [rng.right() for rng in selection]
        )

    def select(self, row_1, col_1, row_2, col_2):
        selection = QItemSelection()
        start_index = self.model().index(row_1, col_1)
        end_index = self.model().index(row_2, col_2)
        selection.select(start_index, end_index)
        self.selectionModel().select(selection, QItemSelectionModel.Select)

    def model(self) -> ParticipantTableModel:
        return super().model()

    def setModel(self, model: ParticipantTableModel):
        if self.model():
            self.model().updated.disconnect()

        super().setModel(model)
        self.model().updated.connect(self.refresh_participant_table)

    def create_context_menu(self):
        menu = QMenu()
        add_action = menu.addAction(QIcon.fromTheme("list-add"), "Add participant")
        add_action.triggered.connect(self.on_add_action_triggered)

        if not self.selectionModel().selection().isEmpty():
            menu.addAction(self.remove_action)
            menu.addSeparator()

            move_up = menu.addAction(
                QIcon.fromTheme("go-up"), "Move selected participants up"
            )
            move_up.triggered.connect(self.on_move_up_action_triggered)
            move_down = menu.addAction(
                QIcon.fromTheme("go-down"), "Move selected participants down"
            )
            move_down.triggered.connect(self.on_move_down_action_triggered)

        return menu

    def contextMenuEvent(self, event: QContextMenuEvent):
        self.create_context_menu().exec_(self.mapToGlobal(event.pos()))

    def refresh_participant_table(self):
        n = len(self.model().participants)
        items = [str(i) for i in range(n)]
        if len(items) >= 2:
            items[0] += " (low)"
            items[-1] += " (high)"

        for row in range(n):
            self.closePersistentEditor(self.model().index(row, 3))

        self.setItemDelegateForColumn(
            2,
            ComboBoxDelegate(
                [""] * len(settings.PARTICIPANT_COLORS),
                colors=settings.PARTICIPANT_COLORS,
                parent=self,
            ),
        )
        self.setItemDelegateForColumn(3, ComboBoxDelegate(items, parent=self))

        for row in range(n):
            self.openPersistentEditor(self.model().index(row, 2))
            self.openPersistentEditor(self.model().index(row, 3))

    @pyqtSlot()
    def on_remove_action_triggered(self):
        self.model().remove_participants(self.selectionModel().selection())

    @pyqtSlot()
    def on_add_action_triggered(self):
        self.model().add_participant()

    @pyqtSlot()
    def on_move_up_action_triggered(self):
        col_start, col_end = self.selected_columns
        start, end = self.model().move_up(self.selectionModel().selection())
        if start is not None and end is not None:
            self.select(start - 1, col_start, end - 1, col_end)

    @pyqtSlot()
    def on_move_down_action_triggered(self):
        col_start, col_end = self.selected_columns
        start, end = self.model().move_down(self.selectionModel().selection())
        if start is not None and end is not None:
            self.select(start + 1, col_start, end + 1, col_end)



from PyQt5.QtCore import pyqtSlot
from PyQt5.QtGui import QCloseEvent

from urh.controller.dialogs.SendDialog import SendDialog
from urh.controller.dialogs.SendRecvDialog import SendRecvDialog
from urh.dev.VirtualDevice import VirtualDevice, Mode
from urh.signalprocessing.ContinuousModulator import ContinuousModulator
from urh.ui.painting.ContinuousSceneManager import ContinuousSceneManager


class ContinuousSendDialog(SendDialog):
    def __init__(
        self,
        project_manager,
        messages,
        modulators,
        total_samples: int,
        parent,
        testing_mode=False,
    ):
        super().__init__(
            project_manager,
            modulated_data=None,
            modulation_msg_indices=None,
            continuous_send_mode=True,
            parent=parent,
            testing_mode=testing_mode,
        )
        self.messages = messages
        self.modulators = modulators

        self.graphics_view = self.ui.graphicsViewContinuousSend
        self.ui.stackedWidget.setCurrentWidget(self.ui.page_continuous_send)
        self.ui.progressBarSample.hide()
        self.ui.lSamplesSentText.hide()

        self.total_samples = total_samples
        self.ui.progressBarMessage.setMaximum(len(messages))

        num_repeats = self.device_settings_widget.ui.spinBoxNRepeat.value()
        self.continuous_modulator = ContinuousModulator(
            messages, modulators, num_repeats=num_repeats
        )
        self.scene_manager = ContinuousSceneManager(
            ring_buffer=self.continuous_modulator.ring_buffer, parent=self
        )
        self.scene_manager.init_scene()
        self.graphics_view.setScene(self.scene_manager.scene)
        self.graphics_view.scene_manager = self.scene_manager

        self.setWindowTitle("Send Signal (continuous mode)")
        self.ui.lSamplesSentText.setText("Progress:")

        self.create_connects()
        self.device_settings_widget.update_for_new_device(overwrite_settings=False)

    def create_connects(self):
        SendRecvDialog.create_connects(self)

    def _update_send_indicator(self, width: int):
        pass

    def update_view(self):
        super().update_view()
        self.ui.progressBarMessage.setValue(
            self.continuous_modulator.current_message_index.value + 1
        )
        self.scene_manager.init_scene()
        self.scene_manager.show_full_scene()
        self.graphics_view.update()

    def closeEvent(self, event: QCloseEvent):
        self.continuous_modulator.stop()
        super().closeEvent(event)

    @pyqtSlot()
    def on_device_started(self):
        super().on_device_started()

    @pyqtSlot()
    def on_device_stopped(self):
        super().on_device_stopped()
        self.continuous_modulator.stop(clear_buffer=False)

    @pyqtSlot()
    def on_stop_clicked(self):
        super().on_stop_clicked()
        self.continuous_modulator.stop()
        self.continuous_modulator.current_message_index.value = 0
        self.scene_manager.clear_path()

    @pyqtSlot()
    def on_start_clicked(self):
        self.device_settings_widget.ui.spinBoxNRepeat.editingFinished.emit()  # inform continuous modulator
        if not self.continuous_modulator.is_running:
            self.continuous_modulator.start()
        super().on_start_clicked()

    @pyqtSlot()
    def on_clear_clicked(self):
        self.continuous_modulator.stop()
        self.continuous_modulator.current_message_index.value = 0
        self.scene_manager.clear_path()
        self.reset()

    @pyqtSlot()
    def on_num_repeats_changed(self):
        super().on_num_repeats_changed()
        self.continuous_modulator.num_repeats = (
            self.device_settings_widget.ui.spinBoxNRepeat.value()
        )

    def on_selected_device_changed(self):
        self.ui.txtEditErrors.clear()
        super().on_selected_device_changed()

    def init_device(self):
        device_name = self.selected_device_name
        num_repeats = self.device_settings_widget.ui.spinBoxNRepeat.value()

        self.device = VirtualDevice(
            self.backend_handler,
            device_name,
            Mode.send,
            device_ip="192.168.10.2",
            sending_repeats=num_repeats,
            parent=self,
        )
        self.ui.btnStart.setEnabled(True)

        try:
            self.device.is_send_continuous = True
            self.device.continuous_send_ring_buffer = (
                self.continuous_modulator.ring_buffer
            )
            self.device.num_samples_to_send = self.total_samples

            self._create_device_connects()
        except ValueError as e:
            self.ui.txtEditErrors.setText("<font color='red'>" + str(e) + "<font>")
            self.ui.btnStart.setEnabled(False)



import array
import copy
from collections import OrderedDict

from PyQt5.QtCore import pyqtSlot, QAbstractTableModel, QModelIndex, Qt, QRegExp
from PyQt5.QtGui import QRegExpValidator
from PyQt5.QtWidgets import QWidget, QHeaderView, QAbstractItemView, QLineEdit

from urh.signalprocessing.ChecksumLabel import ChecksumLabel
from urh.signalprocessing.Message import Message
from urh.ui.delegates.SpinBoxDelegate import SpinBoxDelegate
from urh.ui.ui_checksum_options_widget import Ui_ChecksumOptions
from urh.util import util
from urh.util.GenericCRC import GenericCRC
from urh.util.Logger import logger
from urh.util.WSPChecksum import WSPChecksum


class ChecksumWidget(QWidget):
    SPECIAL_CRCS = OrderedDict(
        [
            ("CC1101", GenericCRC(polynomial="16_standard", start_value=True)),
        ]
    )

    class RangeTableModel(QAbstractTableModel):
        header_labels = ["Start", "End"]

        def __init__(
            self,
            checksum_label: ChecksumLabel,
            message: Message,
            proto_view: int,
            parent=None,
        ):
            """

            :param message:
            :type field_types: list of FieldType
            :param parent:
            """
            super().__init__(parent)
            self.checksum_label = checksum_label
            self.message = message
            self.proto_view = proto_view
            self.update()

        def update(self):
            self.beginResetModel()
            self.endResetModel()

        def columnCount(self, parent: QModelIndex = None, *args, **kwargs):
            return len(self.header_labels)

        def rowCount(self, parent: QModelIndex = None, *args, **kwargs):
            return len(self.checksum_label.data_ranges)

        def headerData(self, section, orientation, role=Qt.DisplayRole):
            if role == Qt.DisplayRole and orientation == Qt.Horizontal:
                return self.header_labels[section]
            return super().headerData(section, orientation, role)

        def data(self, index: QModelIndex, role=Qt.DisplayRole):
            if not index.isValid():
                return None

            i, j = index.row(), index.column()

            if role == Qt.DisplayRole:
                data_range = self.checksum_label.data_ranges[i]
                if j == 0:
                    return (
                        self.message.convert_index(
                            data_range[0], 0, self.proto_view, True
                        )[0]
                        + 1
                    )
                elif j == 1:
                    return self.message.convert_index(
                        data_range[1], 0, self.proto_view, True
                    )[0]
            return None

        def setData(self, index: QModelIndex, value, role: int = ...):
            try:
                int_val = int(value)
            except ValueError:
                return False

            i, j = index.row(), index.column()

            if i > len(self.checksum_label.data_ranges):
                return False

            data_range = self.checksum_label.data_ranges[i]

            if j == 0:
                converted_index = self.message.convert_index(
                    int_val - 1, self.proto_view, 0, True
                )[0]
                if converted_index < data_range[1]:
                    data_range[0] = converted_index
            elif j == 1:
                converted_index = self.message.convert_index(
                    int_val, self.proto_view, 0, True
                )[0]
                if converted_index > data_range[0]:
                    data_range[1] = converted_index

            return True

        def flags(self, index):
            if not index.isValid():
                return Qt.NoItemFlags

            try:
                _ = self.checksum_label.data_ranges[index.row()]
            except IndexError:
                return Qt.NoItemFlags

            return Qt.ItemIsEditable | Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def __init__(
        self,
        checksum_label: ChecksumLabel,
        message: Message,
        proto_view: int,
        parent=None,
    ):
        super().__init__(parent)
        self.ui = Ui_ChecksumOptions()
        self.ui.setupUi(self)
        self.checksum_label = checksum_label
        self.data_range_table_model = self.RangeTableModel(
            checksum_label, message, proto_view, parent=self
        )
        self.ui.tableViewDataRanges.setItemDelegateForColumn(
            0, SpinBoxDelegate(1, 999999, self)
        )
        self.ui.tableViewDataRanges.setItemDelegateForColumn(
            1, SpinBoxDelegate(1, 999999, self)
        )
        self.ui.tableViewDataRanges.horizontalHeader().setSectionResizeMode(
            QHeaderView.Stretch
        )
        self.ui.tableViewDataRanges.setModel(self.data_range_table_model)
        self.ui.tableViewDataRanges.setEditTriggers(QAbstractItemView.AllEditTriggers)
        self.display_crc_data_ranges_in_table()
        self.ui.comboBoxCRCFunction.addItems(
            [crc_name for crc_name in GenericCRC.DEFAULT_POLYNOMIALS]
        )
        self.ui.comboBoxCRCFunction.addItems(
            [special_crc_name for special_crc_name in self.SPECIAL_CRCS]
        )
        self.ui.lineEditCRCPolynomial.setValidator(
            QRegExpValidator(QRegExp("[0-9,a-f]*"))
        )
        self.ui.comboBoxCategory.clear()
        for _, member in self.checksum_label.Category.__members__.items():
            self.ui.comboBoxCategory.addItem(member.value)
        self.set_ui_for_category()
        self.setFocus()
        self.create_connects()

    @property
    def proto_view(self):
        return self.data_range_table_model.proto_view

    @proto_view.setter
    def proto_view(self, value):
        if value != self.data_range_table_model.proto_view:
            self.data_range_table_model.proto_view = value
            self.data_range_table_model.update()

    def create_connects(self):
        self.ui.comboBoxCRCFunction.currentIndexChanged.connect(
            self.on_combobox_crc_function_current_index_changed
        )
        self.ui.btnAddRange.clicked.connect(self.on_btn_add_range_clicked)
        self.ui.btnRemoveRange.clicked.connect(self.on_btn_remove_range_clicked)
        self.ui.lineEditCRCPolynomial.editingFinished.connect(
            self.on_line_edit_crc_polynomial_editing_finished
        )
        self.ui.lineEditStartValue.editingFinished.connect(
            self.on_line_edit_start_value_editing_finished
        )
        self.ui.lineEditFinalXOR.editingFinished.connect(
            self.on_line_edit_final_xor_editing_finished
        )
        self.ui.comboBoxCategory.currentIndexChanged.connect(
            self.on_combobox_category_current_index_changed
        )
        self.ui.radioButtonWSPAuto.clicked.connect(
            self.on_radio_button_wsp_auto_clicked
        )
        self.ui.radioButtonWSPChecksum4.clicked.connect(
            self.on_radio_button_wsp_checksum4_clicked
        )
        self.ui.radioButtonWSPChecksum8.clicked.connect(
            self.on_radio_button_wsp_checksum8_clicked
        )
        self.ui.radioButtonWSPCRC8.clicked.connect(
            self.on_radio_button_wsp_crc8_clicked
        )
        self.ui.checkBoxRefIn.clicked.connect(self.on_check_box_ref_in_clicked)
        self.ui.checkBoxRefOut.clicked.connect(self.on_check_box_ref_out_clicked)

    def set_checksum_ui_elements(self):
        if self.checksum_label.is_generic_crc:
            self.ui.lineEditCRCPolynomial.setText(
                self.checksum_label.checksum.polynomial_as_hex_str
            )
            self.ui.lineEditStartValue.setText(
                util.bit2hex(self.checksum_label.checksum.start_value)
            )
            self.ui.lineEditFinalXOR.setText(
                util.bit2hex(self.checksum_label.checksum.final_xor)
            )
            self.ui.checkBoxRefIn.setChecked(self.checksum_label.checksum.lsb_first)
            self.ui.checkBoxRefOut.setChecked(self.checksum_label.checksum.reverse_all)
            self.__set_crc_function_index()
            self.__ensure_same_length()
            self.__set_crc_info_label()
        elif self.checksum_label.category == self.checksum_label.Category.wsp:
            if self.checksum_label.checksum.mode == WSPChecksum.ChecksumMode.auto:
                self.ui.radioButtonWSPAuto.setChecked(True)
            elif (
                self.checksum_label.checksum.mode == WSPChecksum.ChecksumMode.checksum4
            ):
                self.ui.radioButtonWSPChecksum4.setChecked(True)
            elif (
                self.checksum_label.checksum.mode == WSPChecksum.ChecksumMode.checksum8
            ):
                self.ui.radioButtonWSPChecksum8.setChecked(True)
            elif self.checksum_label.checksum.mode == WSPChecksum.ChecksumMode.crc8:
                self.ui.radioButtonWSPCRC8.setChecked(True)

    def set_ui_for_category(self):
        self.ui.comboBoxCategory.setCurrentText(self.checksum_label.category.value)
        if self.checksum_label.category == self.checksum_label.Category.generic:
            self.ui.stackedWidget.setCurrentWidget(self.ui.page_crc)
        elif self.checksum_label.category == self.checksum_label.Category.wsp:
            self.ui.stackedWidget.setCurrentWidget(self.ui.page_wsp)
        else:
            raise ValueError("Unknown category")

        self.set_checksum_ui_elements()

    def display_crc_data_ranges_in_table(self):
        self.data_range_table_model.update()

    def __set_crc_function_index(self):
        # Get the combobox index
        crc_found = False
        for crc_name in GenericCRC.DEFAULT_POLYNOMIALS:
            test_crc = GenericCRC(crc_name)
            if test_crc == self.checksum_label.checksum:
                self.ui.comboBoxCRCFunction.setCurrentText(crc_name)
                crc_found = True
                break

        if not crc_found:
            for crc_name, crc in self.SPECIAL_CRCS.items():
                if self.checksum_label.checksum == crc:
                    self.ui.comboBoxCRCFunction.setCurrentText(crc_name)
                    crc_found = True
                    break

        if not crc_found:
            self.__add_and_select_custom_item()
        elif "Custom" in [
            self.ui.comboBoxCRCFunction.itemText(i)
            for i in range(self.ui.comboBoxCRCFunction.count())
        ]:
            self.ui.comboBoxCRCFunction.removeItem(
                self.ui.comboBoxCRCFunction.count() - 1
            )

    def __set_crc_info_label(self):
        crc = self.checksum_label.checksum  # type: GenericCRC
        self.ui.label_crc_info.setText(
            "<b>CRC Summary:</b><ul>"
            "<li>Polynomial = {}<>"
            "<li>Length of checksum = {} bit</li>"
            "<li>start value length = {} bit</li>"
            "<li>final XOR length = {} bit</li>"
            "</ul>".format(
                crc.polynomial_to_html,
                crc.poly_order - 1,
                len(crc.start_value),
                len(crc.final_xor),
            )
        )

    def __ensure_same_length(self):
        for dependant_line_edit in [
            self.ui.lineEditStartValue,
            self.ui.lineEditFinalXOR,
        ]:  # type: QLineEdit
            if len(self.ui.lineEditCRCPolynomial.text()) < len(
                dependant_line_edit.text()
            ):
                dependant_line_edit.setText(
                    dependant_line_edit.text()[
                        : len(self.ui.lineEditCRCPolynomial.text())
                    ]
                )
                dependant_line_edit.editingFinished.emit()
            elif len(self.ui.lineEditCRCPolynomial.text()) > len(
                dependant_line_edit.text()
            ):
                # pad zeros at front
                dependant_line_edit.setText(
                    "0"
                    * (
                        len(self.ui.lineEditCRCPolynomial.text())
                        - len(dependant_line_edit.text())
                    )
                    + dependant_line_edit.text()
                )
                dependant_line_edit.editingFinished.emit()

    def __add_and_select_custom_item(self):
        if "Custom" not in [
            self.ui.comboBoxCRCFunction.itemText(i)
            for i in range(self.ui.comboBoxCRCFunction.count())
        ]:
            self.ui.comboBoxCRCFunction.addItem("Custom")
        self.ui.comboBoxCRCFunction.blockSignals(True)
        self.ui.comboBoxCRCFunction.setCurrentText("Custom")
        self.ui.comboBoxCRCFunction.blockSignals(False)

    @pyqtSlot()
    def on_btn_add_range_clicked(self):
        self.checksum_label.data_ranges.append([0, self.checksum_label.start])
        self.data_range_table_model.update()

    @pyqtSlot()
    def on_btn_remove_range_clicked(self):
        if len(self.checksum_label.data_ranges) > 1:
            self.checksum_label.data_ranges.pop(-1)
            self.data_range_table_model.update()

    @pyqtSlot(int)
    def on_combobox_crc_function_current_index_changed(self, index: int):
        poly_str = self.ui.comboBoxCRCFunction.itemText(index)
        if poly_str in GenericCRC.DEFAULT_POLYNOMIALS:
            self.checksum_label.checksum.polynomial = (
                self.checksum_label.checksum.choose_polynomial(poly_str)
            )
            self.checksum_label.checksum.start_value = array.array(
                "B", [0] * (self.checksum_label.checksum.poly_order - 1)
            )
            self.checksum_label.checksum.final_xor = array.array(
                "B", [0] * (self.checksum_label.checksum.poly_order - 1)
            )
        elif poly_str in self.SPECIAL_CRCS:
            self.checksum_label.checksum = copy.deepcopy(self.SPECIAL_CRCS[poly_str])
        else:
            logger.error("Unknown CRC")
            return

        self.ui.lineEditCRCPolynomial.setText(
            self.checksum_label.checksum.polynomial_as_hex_str
        )
        self.ui.lineEditStartValue.setText(
            util.bit2hex(self.checksum_label.checksum.start_value)
        )
        self.ui.lineEditFinalXOR.setText(
            util.bit2hex(self.checksum_label.checksum.final_xor)
        )
        self.ui.lineEditCRCPolynomial.editingFinished.emit()

    @pyqtSlot()
    def on_line_edit_crc_polynomial_editing_finished(self):
        self.checksum_label.checksum.set_polynomial_from_hex(
            self.ui.lineEditCRCPolynomial.text()
        )
        self.__ensure_same_length()
        self.__set_crc_info_label()
        self.__set_crc_function_index()

    @pyqtSlot()
    def on_check_box_ref_in_clicked(self):
        self.checksum_label.checksum.lsb_first = self.ui.checkBoxRefIn.isChecked()

    @pyqtSlot()
    def on_check_box_ref_out_clicked(self):
        self.checksum_label.checksum.reverse_all = self.ui.checkBoxRefOut.isChecked()

    @pyqtSlot()
    def on_line_edit_start_value_editing_finished(self):
        crc = self.checksum_label.checksum
        start_value = util.hex2bit(self.ui.lineEditStartValue.text())
        # pad with zeros at front
        start_value = (
            array.array("B", [0] * (crc.poly_order - 1 - len(start_value)))
            + start_value
        )
        crc.start_value = start_value[0 : crc.poly_order - 1]
        self.ui.lineEditStartValue.setText(util.bit2hex(crc.start_value))
        self.__set_crc_info_label()
        self.__set_crc_function_index()

    @pyqtSlot()
    def on_line_edit_final_xor_editing_finished(self):
        crc = self.checksum_label.checksum
        final_xor = util.hex2bit(self.ui.lineEditFinalXOR.text())
        final_xor = (
            array.array("B", [0] * (crc.poly_order - 1 - len(final_xor))) + final_xor
        )
        crc.final_xor = final_xor[0 : crc.poly_order - 1]
        self.ui.lineEditFinalXOR.setText(util.bit2hex(crc.final_xor))
        self.__set_crc_info_label()
        self.__set_crc_function_index()

    @pyqtSlot(int)
    def on_combobox_category_current_index_changed(self, index: int):
        self.checksum_label.category = self.checksum_label.Category(
            self.ui.comboBoxCategory.currentText()
        )
        self.set_ui_for_category()

    @pyqtSlot()
    def on_radio_button_wsp_auto_clicked(self):
        self.checksum_label.checksum.mode = WSPChecksum.ChecksumMode.auto

    @pyqtSlot()
    def on_radio_button_wsp_checksum4_clicked(self):
        self.checksum_label.checksum.mode = WSPChecksum.ChecksumMode.checksum4

    @pyqtSlot()
    def on_radio_button_wsp_checksum8_clicked(self):
        self.checksum_label.checksum.mode = WSPChecksum.ChecksumMode.checksum8

    @pyqtSlot()
    def on_radio_button_wsp_crc8_clicked(self):
        self.checksum_label.checksum.mode = WSPChecksum.ChecksumMode.crc8
